"""Plotting GL checks."""

from __future__ import annotations

from pyvista.plotting import _vtk


def check_depth_peeling(number_of_peels=100, occlusion_ratio=0.0):
    """Check if depth peeling is available.

    Attempts to use depth peeling to see if it is available for the
    current environment. Returns ``True`` if depth peeling is
    available and has been successfully leveraged, otherwise
    ``False``.

    Parameters
    ----------
    number_of_peels : int, default: 100
        Maximum number of depth peels.

    occlusion_ratio : float, default: 0.0
        Occlusion ratio.

    Returns
    -------
    bool
        ``True`` when system supports depth peeling with the specified
        settings.

    """
    # Try Depth Peeling with a basic scene
    source = _vtk.vtkSphereSource()
    mapper = _vtk.vtkPolyDataMapper()
    mapper.SetInputConnection(source.GetOutputPort())
    actor = _vtk.vtkActor()
    actor.SetMapper(mapper)
    # requires opacity < 1
    actor.GetProperty().SetOpacity(0.5)
    renderer = _vtk.vtkRenderer()
    renderWindow = _vtk.vtkRenderWindow()
    renderWindow.AddRenderer(renderer)
    renderWindow.SetOffScreenRendering(True)
    renderWindow.SetAlphaBitPlanes(True)
    renderWindow.SetMultiSamples(0)
    renderer.AddActor(actor)
    renderer.SetUseDepthPeeling(True)
    renderer.SetMaximumNumberOfPeels(number_of_peels)
    renderer.SetOcclusionRatio(occlusion_ratio)
    renderWindow.Render()
    return renderer.GetLastRenderingUsedDepthPeeling() == 1


def uses_egl() -> bool:
    """Check if VTK has been compiled with EGL support via OSMesa.

    Returns
    -------
    bool
        ``True`` if VTK has been compiled with EGL support via OSMesa,
        otherwise ``False``.

    """
    ren_win_str = str(type(_vtk.vtkRenderWindow()))
    return 'EGL' in ren_win_str or 'OSOpenGL' in ren_win_str



"""Command line initialization.

The imports are needed to initialize the various subcommands (ie. plot, report, convert ...)
"""

# Not sorting imports because it impacts the order commands appear in the `--help` page output.
# ruff: noqa: I001
from __future__ import annotations

from .app import app as app
from . import report as report
from . import convert as convert
from . import plot as plot



"""Affine widget module."""

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import cast

import numpy as np

import pyvista as pv
from pyvista._deprecate_positional_args import _deprecate_positional_args
from pyvista.core.utilities.misc import _NoNewAttrMixin
from pyvista.core.utilities.misc import try_callback

from . import _vtk

if TYPE_CHECKING:
    from pyvista import Actor

DARK_YELLOW = (0.9647058823529412, 0.7450980392156863, 0)
GLOBAL_AXES = np.eye(3)


def _validate_axes(axes):
    """Validate and normalize input axes.

    Axes are expected to follow the right-hand rule (e.g. third axis is the
    cross product of the first two.

    Parameters
    ----------
    axes : sequence
        The axes to be validated and normalized. Should be of shape (3, 3).

    Returns
    -------
    dict
        The validated and normalized axes.

    """
    axes = np.array(axes)
    if axes.shape != (3, 3):
        msg = '`axes` must be a (3, 3) array.'
        raise ValueError(msg)

    axes = axes / np.linalg.norm(axes, axis=1, keepdims=True)
    if not np.allclose(np.cross(axes[0], axes[1]), axes[2]):
        msg = '`axes` do not follow the right hand rule.'
        raise ValueError(msg)

    return axes


def _check_callable(func, name='callback'):
    """Check if a variable is callable."""
    if func and not callable(func):
        msg = f'`{name}` must be a callable, not {type(func)}.'
        raise TypeError(msg)
    return func


def _make_quarter_arc():
    """Make a quarter circle centered at the origin."""
    circ = pv.Circle(resolution=100)
    circ.faces = np.empty(0, dtype=int)
    circ.lines = np.hstack(([26], np.arange(0, 26)))
    return circ


def get_angle(v1, v2):
    """Compute the angle between two vectors in degrees.

    Parameters
    ----------
    v1 : numpy.ndarray
        First input vector.
    v2 : numpy.ndarray
        Second input vector.

    Returns
    -------
    float
        Angle between vectors in degrees.

    """
    return np.rad2deg(np.arccos(np.clip(np.dot(v1, v2), -1.0, 1.0)))


@_deprecate_positional_args
def ray_plane_intersection(start_point, direction, plane_point, normal):  # noqa: PLR0917
    """Compute the intersection between a ray and a plane.

    Parameters
    ----------
    start_point : ndarray
        Starting point of the ray.
    direction : ndarray
        Direction of the ray.
    plane_point : ndarray
        A point on the plane.
    normal : ndarray
        Normal to the plane.

    Returns
    -------
    ndarray
        Intersection point.

    """
    t_value = np.dot(normal, (plane_point - start_point)) / np.dot(normal, direction)
    return start_point + t_value * direction


class AffineWidget3D(_NoNewAttrMixin):
    """3D affine transform widget.

    This widget allows interactive transformations including translation and
    rotation using the left mouse button.

    Parameters
    ----------
    plotter : pyvista.Plotter
        The plotter object.
    actor : pyvista.Actor
        The actor to which the widget is attached to.
    origin : sequence[float], optional
        Origin of the widget. Default is the center of the main actor.
    start : bool, default: True
        If True, start the widget immediately.
    scale : float, default: 0.15
        Scale factor for the widget relative to the length of the actor.
    line_radius : float, default: 0.02
        Relative radius of the lines composing the widget.
    always_visible : bool, default: True
        Make the widget always visible. Setting this to ``False`` will cause
        the widget geometry to be hidden by other actors in the plotter.
    axes_colors : tuple[ColorLike], optional
        Uses the theme by default. Configure the individual axis colors by
        modifying either the theme with ``pyvista.global_theme.axes.x_color =
        <COLOR>`` or setting this with a ``tuple`` as in ``('r', 'g', 'b')``.
    axes : numpy.ndarray, optional
        ``(3, 3)`` Numpy array defining the X, Y, and Z axes. By default this
        matches the default coordinate system.
    release_callback : callable, optional
        Call this method when releasing the left mouse button. It is passed the
        ``user_matrix`` of the actor.
    interact_callback : callable, optional
        Call this method when moving the mouse with the left mouse button
        pressed down and a valid movement actor selected. It is passed the
        ``user_matrix`` of the actor.

    Notes
    -----
    After interacting with the actor, the transform will be stored within
    :attr:`pyvista.Prop3D.user_matrix` but will not be applied to the
    dataset. Use this matrix in conjunction with
    :func:`pyvista.DataObjectFilters.transform` to transform the dataset.

    Requires VTK >= v9.2

    Examples
    --------
    Create the affine widget outside of the plotter and add it.

    >>> import pyvista as pv
    >>> pl = pv.Plotter()
    >>> actor = pl.add_mesh(pv.Sphere())
    >>> widget = pv.AffineWidget3D(pl, actor)
    >>> pl.show()

    Access the transform from the actor.

    >>> actor.user_matrix
    array([[1., 0., 0., 0.],
           [0., 1., 0., 0.],
           [0., 0., 1., 0.],
           [0., 0., 0., 1.]])

    """

    @_deprecate_positional_args(allowed=['plotter', 'actor'])
    def __init__(  # noqa: PLR0917
        self,
        plotter,
        actor,
        origin=None,
        start: bool = True,  # noqa: FBT001, FBT002
        scale=0.15,
        line_radius=0.02,
        always_visible: bool = True,  # noqa: FBT001, FBT002
        axes_colors=None,
        axes=None,
        release_callback=None,
        interact_callback=None,
    ):
        """Initialize the widget."""
        self._axes = np.eye(4)
        self._axes_inv = np.eye(4)
        self._pl = plotter
        self._main_actor = actor
        self._selected_actor: Actor | None = None
        self._init_position = None
        self._mouse_move_observer: int | None = None
        self._left_press_observer: int | None = None
        self._left_release_observer: int | None = None

        if self._main_actor.user_matrix is None:
            self._main_actor.user_matrix = np.eye(4)
        self._cached_matrix = self._main_actor.user_matrix

        self._arrows = []  # type: ignore[var-annotated]
        self._circles = []  # type: ignore[var-annotated]
        self._pressing_down = False
        origin = origin or actor.center
        self._origin = np.array(origin)
        if axes_colors is None:
            axes_colors = (
                pv.global_theme.axes.x_color,
                pv.global_theme.axes.y_color,
                pv.global_theme.axes.z_color,
            )
        self._axes_colors = axes_colors
        self._circ = _make_quarter_arc()
        self._actor_length = self._main_actor.GetLength()
        self._line_radius = line_radius
        self._user_interact_callback = _check_callable(interact_callback)
        self._user_release_callback = _check_callable(release_callback)

        self._init_actors(scale, always_visible)

        # axes must be set after initializing actors
        if axes is not None:
            try:
                _validate_axes(axes)
            except ValueError:
                for actor_ in self._arrows + self._circles:
                    self._pl.remove_actor(actor_)
                raise
            self.axes = axes

        if start:
            self.enable()

    def _init_actors(self, scale, always_visible):
        """Initialize the widget's actors."""
        for ii, color in enumerate(self._axes_colors):
            arrow = pv.Arrow(
                start=(0, 0, 0),
                direction=GLOBAL_AXES[ii],
                scale=self._actor_length * scale * 1.15,
                tip_radius=0.05,
                shaft_radius=self._line_radius,
            )
            self._arrows.append(
                self._pl.add_mesh(arrow, color=color, lighting=False, render=False)
            )
            axis_circ = self._circ.copy()
            if ii == 0:
                axis_circ = axis_circ.rotate_y(-90)
            elif ii == 1:
                axis_circ = axis_circ.rotate_x(90)
            axis_circ.points *= self._main_actor.GetLength() * (scale * 1.6)
            # axis_circ.points += self._origin
            axis_circ = axis_circ.tube(
                radius=self._line_radius * self._actor_length * scale,
                absolute=True,
                radius_factor=1.0,
            )

            self._circles.append(
                self._pl.add_mesh(
                    axis_circ,
                    color=color,
                    lighting=False,
                    render_lines_as_tubes=True,
                    render=False,
                ),
            )

        # update origin and assign a default user_matrix
        for actor in self._arrows + self._circles:
            matrix = np.eye(4)
            matrix[:3, -1] = self._origin
            actor.user_matrix = matrix

        if always_visible:
            for actor in self._arrows + self._circles:
                actor.mapper.SetResolveCoincidentTopologyToPolygonOffset()
                actor.mapper.SetRelativeCoincidentTopologyPolygonOffsetParameters(0, -20000)

    def _get_world_coord_rot(self, interactor):
        """Get the world coordinates given an interactor.

        Unlike ``_get_world_coord_trans``, these coordinates are physically
        accurate, but sensitive to the position of the camera. Rotation is zoom
        independent.

        """
        x, y = interactor.GetEventPosition()
        coordinate = _vtk.vtkCoordinate()
        coordinate.SetCoordinateSystemToDisplay()
        coordinate.SetValue(x, y, 0)
        ren = interactor.GetRenderWindow().GetRenderers().GetFirstRenderer()
        point = np.array(coordinate.GetComputedWorldValue(ren))
        if self._selected_actor:
            index = self._circles.index(self._selected_actor)
            to_widget = np.array(ren.camera.position - self._origin)
            point = ray_plane_intersection(
                start_point=point,
                direction=to_widget,
                plane_point=self._origin,
                normal=self.axes[index],
            )
        return point

    def _get_world_coord_trans(self, interactor):
        """Get the world coordinates given an interactor.

        This uses a modified scaled approach to get the world coordinates that
        are not physically accurate, but ignores zoom and works for
        translation.

        """
        x, y = interactor.GetEventPosition()
        ren = interactor.GetRenderWindow().GetRenderers().GetFirstRenderer()

        # Get normalized view coordinates (-1, 1)
        width, height = ren.GetSize()
        ndc_x = 2 * (x / width) - 1
        ndc_y = 2 * (y / height) - 1
        ndc_z = 1

        # convert camera coordinates to world coordinates
        camera = ren.GetActiveCamera()
        projection_matrix = pv.array_from_vtkmatrix(
            camera.GetProjectionTransformMatrix(ren.GetTiledAspectRatio(), 0, 1),
        )
        inverse_projection_matrix = np.linalg.inv(projection_matrix)
        camera_coords = np.dot(inverse_projection_matrix, [ndc_x, ndc_y, ndc_z, 1])
        modelview_matrix = pv.array_from_vtkmatrix(camera.GetModelViewTransformMatrix())
        inverse_modelview_matrix = np.linalg.inv(modelview_matrix)
        world_coords = np.dot(inverse_modelview_matrix, camera_coords)

        # Scale by twice actor length (experimentally determined for good UX)
        return world_coords[:3] * self._actor_length * 2

    def _move_callback(self, interactor, _event):
        """Process actions for the move mouse event."""
        click_x, click_y = interactor.GetEventPosition()
        click_z = 0
        picker = interactor.GetPicker()
        renderer = interactor.GetInteractorStyle()._parent()._plotter.iren.get_poked_renderer()
        picker.Pick(click_x, click_y, click_z, renderer)
        actor = picker.GetActor()

        if self._pressing_down:
            if self._selected_actor in self._arrows:
                current_pos = self._get_world_coord_trans(interactor)
                index = self._arrows.index(self._selected_actor)
                diff = current_pos - self._init_position
                trans_matrix = np.eye(4)
                trans_matrix[:3, -1] = self.axes[index] * np.dot(diff, self.axes[index])
                matrix = trans_matrix @ self._cached_matrix
            elif self._selected_actor in self._circles:
                current_pos = self._get_world_coord_rot(interactor)
                index = self._circles.index(self._selected_actor)
                vec_current = current_pos - self._origin
                vec_init = self._init_position - self._origin
                normal = self.axes[index]
                vec_current = vec_current - np.dot(vec_current, normal) * normal
                vec_init = vec_init - np.dot(vec_init, normal) * normal
                vec_current /= np.linalg.norm(vec_current)
                vec_init /= np.linalg.norm(vec_init)
                angle = get_angle(vec_init, vec_current)
                cross = np.cross(vec_init, vec_current)
                if cross[index] < 0:
                    angle = -angle

                trans = _vtk.vtkTransform()
                trans.Translate(self._origin)  # type: ignore[call-overload]
                trans.RotateWXYZ(
                    angle,
                    self._axes[index][0],
                    self._axes[index][1],
                    self._axes[index][2],
                )
                trans.Translate(-self._origin)  # type: ignore[call-overload]
                trans.Update()
                rot_matrix = pv.array_from_vtkmatrix(trans.GetMatrix())
                matrix = rot_matrix @ self._cached_matrix

            if self._user_interact_callback:
                try_callback(self._user_interact_callback, self._main_actor.user_matrix)

            self._main_actor.user_matrix = matrix

        elif self._selected_actor and self._selected_actor is not actor:
            # Return the color of the currently selected actor to normal and
            # deselect it
            if self._selected_actor in self._arrows:
                index = self._arrows.index(self._selected_actor)
            elif self._selected_actor in self._circles:
                index = self._circles.index(self._selected_actor)
            self._selected_actor.prop.color = self._axes_colors[index]
            self._selected_actor = None

        # Highlight the actor if there is no selected actor
        if actor and not self._selected_actor:
            if actor in self._arrows:
                index = self._arrows.index(actor)
                self._arrows[index].prop.color = DARK_YELLOW
                actor.prop.color = DARK_YELLOW
                self._selected_actor = actor
            elif actor in self._circles:
                index = self._circles.index(actor)
                self._circles[index].prop.color = DARK_YELLOW
                actor.prop.color = DARK_YELLOW
                self._selected_actor = actor
        self._pl.render()

    def _press_callback(self, interactor, _event):
        """Process actions for the mouse button press event."""
        if self._selected_actor:
            self._pl.enable_trackball_actor_style()
            self._pressing_down = True
            if self._selected_actor in self._circles:
                self._init_position = self._get_world_coord_rot(interactor)
            else:
                self._init_position = self._get_world_coord_trans(interactor)

    def _release_callback(self, _interactor, _event):
        """Process actions for the mouse button release event."""
        self._pl.enable_trackball_style()
        self._pressing_down = False
        self._cached_matrix = self._main_actor.user_matrix
        if self._user_release_callback:
            try_callback(self._user_release_callback, self._main_actor.user_matrix)

    def _reset(self):
        """Reset the actor and cached transform."""
        self._main_actor.user_matrix = np.eye(4)
        self._cached_matrix = np.eye(4)

    @property
    def axes(self):
        """Return or set the axes of the widget.

        The axes will be checked for orthogonality. Non-orthogonal axes will
        raise a ``ValueError``

        Returns
        -------
        numpy.ndarray
            ``(3, 3)`` array of axes.

        """
        return self._axes[:3, :3]

    @axes.setter
    def axes(self, axes):
        mat = np.eye(4)
        mat[:3, :3] = _validate_axes(axes)
        mat[:3, -1] = self.origin
        self._axes = mat
        self._axes_inv = np.linalg.inv(self._axes)  # type: ignore[assignment]
        for actor in self._arrows + self._circles:
            matrix = actor.user_matrix
            # Be sure to use the inverse here
            matrix[:3, :3] = self._axes_inv[:3, :3]
            actor.user_matrix = matrix

    @property
    def origin(self) -> tuple[float, float, float]:
        """Origin of the widget.

        This is where the origin of the widget will be located and where the
        actor will be rotated about.

        Returns
        -------
        tuple
            Widget origin.

        """
        return cast('tuple[float, float, float]', tuple(self._origin))

    @origin.setter
    def origin(self, value):
        value = np.array(value)
        diff = value - self._origin

        for actor in self._circles + self._arrows:
            if actor.user_matrix is None:
                actor.user_matrix = np.eye(4)
            matrix = actor.user_matrix
            matrix[:3, -1] += diff
            actor.user_matrix = matrix

        self._origin = value

    def enable(self):
        """Enable the widget."""
        if not self._pl._picker_in_use:
            self._pl.enable_mesh_picking(show_message=False, show=False, picker='hardware')
        self._mouse_move_observer = self._pl.iren.add_observer(
            'MouseMoveEvent',
            self._move_callback,
        )
        self._left_press_observer = self._pl.iren.add_observer(
            'LeftButtonPressEvent',
            self._press_callback,
            interactor_style_fallback=False,
        )
        self._left_release_observer = self._pl.iren.add_observer(
            'LeftButtonReleaseEvent',
            self._release_callback,
            interactor_style_fallback=False,
        )

    def disable(self):
        """Disable the widget."""
        self._pl.disable_picking()
        if self._mouse_move_observer:
            self._pl.iren.remove_observer(self._mouse_move_observer)
        if self._left_press_observer:
            self._pl.iren.remove_observer(self._left_press_observer)
        if self._left_release_observer:
            self._pl.iren.remove_observer(self._left_release_observer)

    def remove(self):
        """Disable and delete all actors of this widget."""
        self.disable()
        for actor in self._circles + self._arrows:
            self._pl.remove_actor(actor)
        self._circles = []
        self._arrows = []



"""Module with enum options classes for plotting."""

from __future__ import annotations

from pyvista.core.utilities.misc import AnnotatedIntEnum


class InterpolationType(AnnotatedIntEnum):
    """Lighting interpolation types.

    Attributes
    ----------
    FLAT : (int, str)
        Flat interpolation type.
    GOURAUD : (int, str)
        Gouraud interpolation type.
    PHONG : (int, str)
        Phong interpolation type.
    PBR : (int, str)
        Physically based rendering interpolation type.

    """

    FLAT = (0, 'Flat')
    GOURAUD = (1, 'Gouraud')
    PHONG = (2, 'PHONG')
    PBR = (3, 'Physically based rendering')

    @classmethod
    def from_str(cls, input_str):
        """Create from string.

        Create an instance of InterpolationType from a string.

        Parameters
        ----------
        input_str : str
            The string representation of the interpolation type.  Accepts
            aliases such as ``'pbr'`` for ``'Physically based rendering'``.

        Returns
        -------
        InterpolationType
            Interpolation type as defined by the input string.

        """
        aliases = {
            'pbr': 'Physically based rendering',
        }
        if input_str in aliases:
            input_str = aliases[input_str]
        return super().from_str(input_str)


class RepresentationType(AnnotatedIntEnum):
    """Types of representations the models can have."""

    POINTS = (0, 'Points')
    WIREFRAME = (1, 'Wireframe')
    SURFACE = (2, 'Surface')


class ElementType(AnnotatedIntEnum):
    """Types of elemental geometries."""

    MESH = (0, 'Mesh')
    CELL = (1, 'Cell')
    FACE = (2, 'Face')
    EDGE = (3, 'Edge')
    POINT = (4, 'Point')


class PickerType(AnnotatedIntEnum):
    """Types of pickers."""

    AREA = (0, 'Area')
    CELL = (1, 'Cell')
    HARDWARE = (2, 'Hardware')
    POINT = (3, 'Point')
    PROP = (4, 'Prop')
    RENDERED = (5, 'Rendered')
    RESLICE = (6, 'Reslice')
    SCENE = (7, 'Scene')
    VOLUME = (8, 'Volume')
    WORLD = (9, 'World')



"""Test the pre-commit hooks"""

from __future__ import annotations

import shlex
import subprocess
import sys
import textwrap
from typing import TYPE_CHECKING

import pytest
import yaml

if TYPE_CHECKING:
    from pathlib import Path


@pytest.fixture(scope='session')
def pre_commit_config(request: pytest.FixtureRequest):
    with (request.config.rootpath / '.pre-commit-config.yaml').open() as f:
        return yaml.safe_load(f)


def test_warnings_converter(
    tmp_path: Path,
    pre_commit_config: dict,
    request: pytest.FixtureRequest,
):
    test = """\
    import warnings

    warnings.warn("foo")
    warnings.warn("foo", UserWarning)
    warnings.warn("foo", UserWarning, 1)
    warnings.warn("foo", UserWarning, stacklevel=1)
    warnings.warn("foo", category=UserWarning, stacklevel=1)
    warnings.warn(message="foo", category=UserWarning, stacklevel=1)
    warnings.warn(category=UserWarning, stacklevel=1, message="foo")
    """

    if sys.version_info[:2] >= (3, 12):
        test += """
    warnings.warn(category=UserWarning, stacklevel=1, message="foo", source='bar', skip_file_prefixes=('',))
    """  # noqa: E501

    with (file := (tmp_path / 'file.py')).open('w') as f:
        f.write(textwrap.dedent(test))

    local = next(v for v in pre_commit_config['repos'] if v['repo'] == 'local')
    warning_hook = next(v for v in local['hooks'] if v['id'] == 'warn_external')
    cml = warning_hook['entry']

    ret = subprocess.run(
        [sys.executable, *shlex.split(cml)[1:], str(file.absolute())],
        check=True,
        cwd=request.config.rootpath,
    )
    assert ret.returncode == 0

    with file.open('r') as f:
        lines = f.readlines()

    expected = """\
        from pyvista._warn_external import warn_external

        warn_external("foo")
        warn_external("foo", UserWarning)
        warn_external("foo", UserWarning)
        warn_external("foo", UserWarning)
        warn_external("foo", category=UserWarning)
        warn_external(message="foo", category=UserWarning)
        warn_external(message="foo", category=UserWarning)
        """

    if sys.version_info[:2] >= (3, 12):
        expected += """
        warn_external(message="foo", category=UserWarning)
        """

    assert textwrap.dedent(expected) == ''.join(lines)



from __future__ import annotations

import pathlib
import sys
from typing import TYPE_CHECKING

import pytest

from pyvista._warn_external import warn_external

if TYPE_CHECKING:
    from pytest_mock import MockerFixture


def test_warn_external(recwarn: pytest.WarningsRecorder):
    """Taken and adapted from
    https://github.com/matplotlib/matplotlib/blob/a00d606d592bcf8d335f4f3ac2768882d3a49e7b/lib/matplotlib/tests/test_cbook.py#L509"""

    warn_external('oops')
    assert len(recwarn) == 1
    if sys.version_info[:2] >= (3, 12):
        # With Python 3.12, we let Python figure out the stacklevel using the
        # `skip_file_prefixes` argument, which cannot exempt tests, so just confirm
        # the filename is not in the package.
        basedir = pathlib.Path(__file__).parents[1]
        assert not recwarn[0].filename.startswith(str(basedir / 'pyvista'))

    else:
        # On older Python versions, we manually calculated the stacklevel, and had an
        # exception for our own tests.
        assert recwarn[0].filename == __file__


def test_warn_external_frame_embedded_python(mocker: MockerFixture):
    """Taken and adapted from
    https://github.com/matplotlib/matplotlib/blob/a00d606d592bcf8d335f4f3ac2768882d3a49e7b/lib/matplotlib/tests/test_cbook.py#L525"""
    m = mocker.patch.object(sys, '_getframe')
    m.return_value = None
    with pytest.warns(UserWarning, match=r'\Adummy\Z'):
        warn_external('dummy')



"""Color module supporting plotting module.

Used code from matplotlib.colors.  Thanks for your work.
"""

# Necessary for autodoc_type_aliases to recognize the type aliases used in the signatures
# of methods defined in this module.
from __future__ import annotations

from colorsys import rgb_to_hls
import contextlib
import importlib
import inspect
from typing import Literal
from typing import get_args

from cycler import Cycler
from cycler import cycler

try:
    from matplotlib import colormaps
    from matplotlib import colors
except ImportError:  # pragma: no cover
    # typing for newer versions of matplotlib
    # in newer versions cm is a module
    from matplotlib import cm as colormaps  # type: ignore[assignment]
    from matplotlib import colors

from typing import TYPE_CHECKING
from typing import Any

from matplotlib.colors import ListedColormap
import matplotlib.pyplot as plt
import numpy as np

import pyvista as pv
from pyvista import _validation
from pyvista._deprecate_positional_args import _deprecate_positional_args
from pyvista.core.utilities.misc import _NoNewAttrMixin

from . import _vtk

if TYPE_CHECKING:
    from ._typing import ColorLike
    from ._typing import ColormapOptions

IPYGANY_MAP = {
    'reds': 'Reds',
    'spectral': 'Spectral',
}

_ALLOWED_COLOR_NAME_DELIMITERS = '_' + '-' + ' '
_REMOVE_DELIMITER_LOOKUP = str.maketrans('', '', _ALLOWED_COLOR_NAME_DELIMITERS)


def _format_color_name(name: str):
    """Format name as lower-case and remove delimiters."""
    return name.lower().translate(_REMOVE_DELIMITER_LOOKUP)


def _format_color_dict(colors: dict[str, str]):
    """Format name and hex value."""
    return {_format_color_name(n): h.lower() for n, h in colors.items()}


# Colors from the CSS standard. Matches matplotlib.colors.CSS4_COLORS
# but with synonyms removed
_CSS_COLORS = {
    'aliceblue': '#F0F8FF',
    'antiquewhite': '#FAEBD7',
    'aquamarine': '#7FFFD4',
    'azure': '#F0FFFF',
    'beige': '#F5F5DC',
    'bisque': '#FFE4C4',
    'black': '#000000',
    'blanchedalmond': '#FFEBCD',
    'blue': '#0000FF',
    'blueviolet': '#8A2BE2',
    'brown': '#A52A2A',
    'burlywood': '#DEB887',
    'cadetblue': '#5F9EA0',
    'chartreuse': '#7FFF00',
    'chocolate': '#D2691E',
    'coral': '#FF7F50',
    'cornflowerblue': '#6495ED',
    'cornsilk': '#FFF8DC',
    'crimson': '#DC143C',
    'cyan': '#00FFFF',
    'darkblue': '#00008B',
    'darkcyan': '#008B8B',
    'darkgoldenrod': '#B8860B',
    'darkgray': '#A9A9A9',
    'darkgreen': '#006400',
    'darkkhaki': '#BDB76B',
    'darkmagenta': '#8B008B',
    'darkolivegreen': '#556B2F',
    'darkorange': '#FF8C00',
    'darkorchid': '#9932CC',
    'darkred': '#8B0000',
    'darksalmon': '#E9967A',
    'darkseagreen': '#8FBC8F',
    'darkslateblue': '#483D8B',
    'darkslategray': '#2F4F4F',
    'darkturquoise': '#00CED1',
    'darkviolet': '#9400D3',
    'deeppink': '#FF1493',
    'deepskyblue': '#00BFFF',
    'dimgray': '#696969',
    'dodgerblue': '#1E90FF',
    'firebrick': '#B22222',
    'floralwhite': '#FFFAF0',
    'forestgreen': '#228B22',
    'gainsboro': '#DCDCDC',
    'ghostwhite': '#F8F8FF',
    'gold': '#FFD700',
    'goldenrod': '#DAA520',
    'gray': '#808080',
    'green': '#008000',
    'greenyellow': '#ADFF2F',
    'honeydew': '#F0FFF0',
    'hotpink': '#FF69B4',
    'indianred': '#CD5C5C',
    'indigo': '#4B0082',
    'ivory': '#FFFFF0',
    'khaki': '#F0E68C',
    'lavender': '#E6E6FA',
    'lavenderblush': '#FFF0F5',
    'lawngreen': '#7CFC00',
    'lemonchiffon': '#FFFACD',
    'lightblue': '#ADD8E6',
    'lightcoral': '#F08080',
    'lightcyan': '#E0FFFF',
    'lightgoldenrodyellow': '#FAFAD2',
    'lightgray': '#D3D3D3',
    'lightgreen': '#90EE90',
    'lightpink': '#FFB6C1',
    'lightsalmon': '#FFA07A',
    'lightseagreen': '#20B2AA',
    'lightskyblue': '#87CEFA',
    'lightslategray': '#778899',
    'lightsteelblue': '#B0C4DE',
    'lightyellow': '#FFFFE0',
    'lime': '#00FF00',
    'limegreen': '#32CD32',
    'linen': '#FAF0E6',
    'magenta': '#FF00FF',
    'maroon': '#800000',
    'mediumaquamarine': '#66CDAA',
    'mediumblue': '#0000CD',
    'mediumorchid': '#BA55D3',
    'mediumpurple': '#9370DB',
    'mediumseagreen': '#3CB371',
    'mediumslateblue': '#7B68EE',
    'mediumspringgreen': '#00FA9A',
    'mediumturquoise': '#48D1CC',
    'mediumvioletred': '#C71585',
    'midnightblue': '#191970',
    'mintcream': '#F5FFFA',
    'mistyrose': '#FFE4E1',
    'moccasin': '#FFE4B5',
    'navajowhite': '#FFDEAD',
    'navy': '#000080',
    'oldlace': '#FDF5E6',
    'olive': '#808000',
    'olivedrab': '#6B8E23',
    'orange': '#FFA500',
    'orangered': '#FF4500',
    'orchid': '#DA70D6',
    'palegoldenrod': '#EEE8AA',
    'palegreen': '#98FB98',
    'paleturquoise': '#AFEEEE',
    'palevioletred': '#DB7093',
    'papayawhip': '#FFEFD5',
    'peachpuff': '#FFDAB9',
    'peru': '#CD853F',
    'pink': '#FFC0CB',
    'plum': '#DDA0DD',
    'powderblue': '#B0E0E6',
    'purple': '#800080',
    'rebeccapurple': '#663399',
    'red': '#FF0000',
    'rosybrown': '#BC8F8F',
    'royalblue': '#4169E1',
    'saddlebrown': '#8B4513',
    'salmon': '#FA8072',
    'sandybrown': '#F4A460',
    'seagreen': '#2E8B57',
    'seashell': '#FFF5EE',
    'sienna': '#A0522D',
    'silver': '#C0C0C0',
    'skyblue': '#87CEEB',
    'slateblue': '#6A5ACD',
    'slategray': '#708090',
    'snow': '#FFFAFA',
    'springgreen': '#00FF7F',
    'steelblue': '#4682B4',
    'tan': '#D2B48C',
    'teal': '#008080',
    'thistle': '#D8BFD8',
    'tomato': '#FF6347',
    'turquoise': '#40E0D0',
    'violet': '#EE82EE',
    'wheat': '#F5DEB3',
    'white': '#FFFFFF',
    'whitesmoke': '#F5F5F5',
    'yellow': '#FFFF00',
    'yellowgreen': '#9ACD32',
}

# Tableau colors. Matches matplotlib.colors.TABLEAU_COLORS
_TABLEAU_COLORS = {
    'tab:blue': '#1f77b4',
    'tab:orange': '#ff7f0e',
    'tab:green': '#2ca02c',
    'tab:red': '#d62728',
    'tab:purple': '#9467bd',
    'tab:brown': '#8c564b',
    'tab:pink': '#e377c2',
    'tab:gray': '#7f7f7f',
    'tab:olive': '#bcbd22',
    'tab:cyan': '#17becf',
}
_PARAVIEW_COLORS = {'paraview_background': '#52576e'}

# Colors from https://htmlpreview.github.io/?https://github.com/Kitware/vtk-examples/blob/gh-pages/VTKNamedColorPatches.html
# The vtk colors are only partially supported:
# - VTK colors with the same name as CSS colors but different values are excluded
#   (i.e. the CSS colors take precedent)
# - Not all VTK synonyms are supported.
# - Colors with adjective suffixes are renamed to use a prefix instead
#   (e.g. 'green_pale' is renamed to 'pale_green'). This is done to keep VTK color
#   names consistent with CSS names. In most cases this altered color name is
#   supported directly by vtkNamedColors, but in some cases this technically is no
#   longer a valid named vtk color. See tests.
_VTK_COLORS = {
    'alizarin_crimson': '#e32636',
    'aureoline_yellow': '#ffa824',
    'banana': '#e3cf57',
    'brick': '#9c661f',
    'brown_madder': '#db2929',
    'brown_ochre': '#87421f',
    'burnt_sienna': '#8a360f',
    'burnt_umber': '#8a3324',
    'cadmium_lemon': '#ffe303',
    'cadmium_orange': '#ff6103',
    'cadmium_yellow': '#ff9912',
    'carrot': '#ed9121',
    'cerulean': '#05b8cc',
    'chrome_oxide_green': '#668014',
    'cinnabar_green': '#61b329',
    'cobalt': '#3d59ab',
    'cobalt_green': '#3d9140',
    'cold_grey': '#808a87',
    'deep_cadmium_red': '#e3170d',
    'deep_cobalt_violet': '#91219e',
    'deep_naples_yellow': '#ffa812',
    'deep_ochre': '#733d1a',
    'eggshell': '#fce6c9',
    'emerald_green': '#00c957',
    'english_red': '#d43d1a',
    'flesh': '#ff7d40',
    'flesh_ochre': '#ff5721',
    'geranium_lake': '#e31230',
    'gold_ochre': '#c77826',
    'greenish_umber': '#ff3d0d',
    'ivory_black': '#292421',
    'lamp_black': '#2e473b',
    'light_cadmium_red': '#ff030d',
    'light_cadmium_yellow': '#ffb00f',
    'light_slate_blue': '#8470ff',
    'light_viridian': '#6eff70',
    'madder_lake_deep': '#e32e30',
    'manganese_blue': '#03a89e',
    'mars_orange': '#964514',
    'mars_yellow': '#e3701a',
    'melon': '#e3a869',
    'mint': '#bdfcc9',
    'peacock': '#33a1c9',
    'permanent_green': '#0ac92b',
    'permanent_red_violet': '#db2645',
    'raspberry': '#872657',
    'raw_sienna': '#C76114',
    'raw_umber': '#734a12',
    'rose_madder': '#e33638',
    'sap_green': '#308014',
    'sepia': '#5e2612',
    'terre_verte': '#385e0f',
    'titanium_white': '#fcfff0',
    'turquoise_blue': '#00c78c',
    'ultramarine': '#120a8f',
    'ultramarine_violet': '#5c246e',
    'van_dyke_brown': '#5e2605',
    'venetian_red': '#d41a1f',
    'violet_red': '#d02090',
    'warm_grey': '#808069',
    'yellow_ochre': '#e38217',
    'zinc_white': '#fcf7ff',
}

hexcolors = _format_color_dict(_CSS_COLORS | _PARAVIEW_COLORS | _TABLEAU_COLORS | _VTK_COLORS)

color_names = {h: n for n, h in hexcolors.items()}

color_char_to_word = {
    'b': 'blue',
    'g': 'green',
    'r': 'red',
    'c': 'cyan',
    'm': 'magenta',
    'y': 'yellow',
    'k': 'black',
    'w': 'white',
}

_color_synonyms = {
    **color_char_to_word,
    'aqua': 'cyan',
    'darkgrey': 'darkgray',
    'darkslategrey': 'darkslategray',
    'dimgrey': 'dimgray',
    'fuchsia': 'magenta',
    'grey': 'gray',
    'lightgrey': 'lightgray',
    'lightslategrey': 'lightslategray',
    'pv': 'paraview_background',
    'paraview': 'paraview_background',
    'slategrey': 'slategray',
    'lightgoldenrod': 'lightgoldenrodyellow',
}

color_synonyms = {
    _format_color_name(syn): _format_color_name(name) for syn, name in _color_synonyms.items()
}

matplotlib_default_colors = [
    '#1f77b4',
    '#ff7f0e',
    '#2ca02c',
    '#d62728',
    '#9467bd',
    '#8c564b',
    '#e377c2',
    '#7f7f7f',
    '#bcbd22',
    '#17becf',
]

COLOR_SCHEMES = {
    'spectrum': {
        'id': _vtk.vtkColorSeries.SPECTRUM,
        'descr': 'black, red, blue, green, purple, orange, brown',
    },
    'warm': {'id': _vtk.vtkColorSeries.WARM, 'descr': 'dark red → yellow'},
    'cool': {'id': _vtk.vtkColorSeries.COOL, 'descr': 'green → blue → purple'},
    'blues': {'id': _vtk.vtkColorSeries.BLUES, 'descr': 'Different shades of blue'},
    'wild_flower': {
        'id': _vtk.vtkColorSeries.WILD_FLOWER,
        'descr': 'blue → purple → pink',
    },
    'citrus': {'id': _vtk.vtkColorSeries.CITRUS, 'descr': 'green → yellow → orange'},
    'div_purple_orange11': {
        'id': _vtk.vtkColorSeries.BREWER_DIVERGING_PURPLE_ORANGE_11,
        'descr': 'dark brown → white → dark purple',
    },
    'div_purple_orange10': {
        'id': _vtk.vtkColorSeries.BREWER_DIVERGING_PURPLE_ORANGE_10,
        'descr': 'dark brown → white → dark purple',
    },
    'div_purple_orange9': {
        'id': _vtk.vtkColorSeries.BREWER_DIVERGING_PURPLE_ORANGE_9,
        'descr': 'brown → white → purple',
    },
    'div_purple_orange8': {
        'id': _vtk.vtkColorSeries.BREWER_DIVERGING_PURPLE_ORANGE_8,
        'descr': 'brown → white → purple',
    },
    'div_purple_orange7': {
        'id': _vtk.vtkColorSeries.BREWER_DIVERGING_PURPLE_ORANGE_7,
        'descr': 'brown → white → purple',
    },
    'div_purple_orange6': {
        'id': _vtk.vtkColorSeries.BREWER_DIVERGING_PURPLE_ORANGE_6,
        'descr': 'brown → white → purple',
    },
    'div_purple_orange5': {
        'id': _vtk.vtkColorSeries.BREWER_DIVERGING_PURPLE_ORANGE_5,
        'descr': 'orange → white → purple',
    },
    'div_purple_orange4': {
        'id': _vtk.vtkColorSeries.BREWER_DIVERGING_PURPLE_ORANGE_4,
        'descr': 'orange → white → purple',
    },
    'div_purple_orange3': {
        'id': _vtk.vtkColorSeries.BREWER_DIVERGING_PURPLE_ORANGE_3,
        'descr': 'orange → white → purple',
    },
    'div_spectral11': {
        'id': _vtk.vtkColorSeries.BREWER_DIVERGING_SPECTRAL_11,
        'descr': 'dark red → light yellow → dark blue',
    },
    'div_spectral10': {
        'id': _vtk.vtkColorSeries.BREWER_DIVERGING_SPECTRAL_10,
        'descr': 'dark red → light yellow → dark blue',
    },
    'div_spectral9': {
        'id': _vtk.vtkColorSeries.BREWER_DIVERGING_SPECTRAL_9,
        'descr': 'red → light yellow → blue',
    },
    'div_spectral8': {
        'id': _vtk.vtkColorSeries.BREWER_DIVERGING_SPECTRAL_8,
        'descr': 'red → light yellow → blue',
    },
    'div_spectral7': {
        'id': _vtk.vtkColorSeries.BREWER_DIVERGING_SPECTRAL_7,
        'descr': 'red → light yellow → blue',
    },
    'div_spectral6': {
        'id': _vtk.vtkColorSeries.BREWER_DIVERGING_SPECTRAL_6,
        'descr': 'red → light yellow → blue',
    },
    'div_spectral5': {
        'id': _vtk.vtkColorSeries.BREWER_DIVERGING_SPECTRAL_5,
        'descr': 'red → light yellow → blue',
    },
    'div_spectral4': {
        'id': _vtk.vtkColorSeries.BREWER_DIVERGING_SPECTRAL_4,
        'descr': 'red → light yellow → blue',
    },
    'div_spectral3': {
        'id': _vtk.vtkColorSeries.BREWER_DIVERGING_SPECTRAL_3,
        'descr': 'orange → light yellow → green',
    },
    'div_brown_blue_green11': {
        'id': _vtk.vtkColorSeries.BREWER_DIVERGING_BROWN_BLUE_GREEN_11,
        'descr': 'dark brown → white → dark blue-green',
    },
    'div_brown_blue_green10': {
        'id': _vtk.vtkColorSeries.BREWER_DIVERGING_BROWN_BLUE_GREEN_10,
        'descr': 'dark brown → white → dark blue-green',
    },
    'div_brown_blue_green9': {
        'id': _vtk.vtkColorSeries.BREWER_DIVERGING_BROWN_BLUE_GREEN_9,
        'descr': 'brown → white → blue-green',
    },
    'div_brown_blue_green8': {
        'id': _vtk.vtkColorSeries.BREWER_DIVERGING_BROWN_BLUE_GREEN_8,
        'descr': 'brown → white → blue-green',
    },
    'div_brown_blue_green7': {
        'id': _vtk.vtkColorSeries.BREWER_DIVERGING_BROWN_BLUE_GREEN_7,
        'descr': 'brown → white → blue-green',
    },
    'div_brown_blue_green6': {
        'id': _vtk.vtkColorSeries.BREWER_DIVERGING_BROWN_BLUE_GREEN_6,
        'descr': 'brown → white → blue-green',
    },
    'div_brown_blue_green5': {
        'id': _vtk.vtkColorSeries.BREWER_DIVERGING_BROWN_BLUE_GREEN_5,
        'descr': 'brown → white → blue-green',
    },
    'div_brown_blue_green4': {
        'id': _vtk.vtkColorSeries.BREWER_DIVERGING_BROWN_BLUE_GREEN_4,
        'descr': 'brown → white → blue-green',
    },
    'div_brown_blue_green3': {
        'id': _vtk.vtkColorSeries.BREWER_DIVERGING_BROWN_BLUE_GREEN_3,
        'descr': 'brown → white → blue-green',
    },
    'seq_blue_green9': {
        'id': _vtk.vtkColorSeries.BREWER_SEQUENTIAL_BLUE_GREEN_9,
        'descr': 'light blue → dark green',
    },
    'seq_blue_green8': {
        'id': _vtk.vtkColorSeries.BREWER_SEQUENTIAL_BLUE_GREEN_8,
        'descr': 'light blue → dark green',
    },
    'seq_blue_green7': {
        'id': _vtk.vtkColorSeries.BREWER_SEQUENTIAL_BLUE_GREEN_7,
        'descr': 'light blue → dark green',
    },
    'seq_blue_green6': {
        'id': _vtk.vtkColorSeries.BREWER_SEQUENTIAL_BLUE_GREEN_6,
        'descr': 'light blue → green',
    },
    'seq_blue_green5': {
        'id': _vtk.vtkColorSeries.BREWER_SEQUENTIAL_BLUE_GREEN_5,
        'descr': 'light blue → green',
    },
    'seq_blue_green4': {
        'id': _vtk.vtkColorSeries.BREWER_SEQUENTIAL_BLUE_GREEN_4,
        'descr': 'light blue → green',
    },
    'seq_blue_green3': {
        'id': _vtk.vtkColorSeries.BREWER_SEQUENTIAL_BLUE_GREEN_3,
        'descr': 'light blue → green',
    },
    'seq_yellow_orange_brown9': {
        'id': _vtk.vtkColorSeries.BREWER_SEQUENTIAL_YELLOW_ORANGE_BROWN_9,
        'descr': 'light yellow → orange → dark brown',
    },
    'seq_yellow_orange_brown8': {
        'id': _vtk.vtkColorSeries.BREWER_SEQUENTIAL_YELLOW_ORANGE_BROWN_8,
        'descr': 'light yellow → orange → brown',
    },
    'seq_yellow_orange_brown7': {
        'id': _vtk.vtkColorSeries.BREWER_SEQUENTIAL_YELLOW_ORANGE_BROWN_7,
        'descr': 'light yellow → orange → brown',
    },
    'seq_yellow_orange_brown6': {
        'id': _vtk.vtkColorSeries.BREWER_SEQUENTIAL_YELLOW_ORANGE_BROWN_6,
        'descr': 'light yellow → orange → brown',
    },
    'seq_yellow_orange_brown5': {
        'id': _vtk.vtkColorSeries.BREWER_SEQUENTIAL_YELLOW_ORANGE_BROWN_5,
        'descr': 'light yellow → orange → brown',
    },
    'seq_yellow_orange_brown4': {
        'id': _vtk.vtkColorSeries.BREWER_SEQUENTIAL_YELLOW_ORANGE_BROWN_4,
        'descr': 'light yellow → orange',
    },
    'seq_yellow_orange_brown3': {
        'id': _vtk.vtkColorSeries.BREWER_SEQUENTIAL_YELLOW_ORANGE_BROWN_3,
        'descr': 'light yellow → orange',
    },
    'seq_blue_purple9': {
        'id': _vtk.vtkColorSeries.BREWER_SEQUENTIAL_BLUE_PURPLE_9,
        'descr': 'light blue → dark purple',
    },
    'seq_blue_purple8': {
        'id': _vtk.vtkColorSeries.BREWER_SEQUENTIAL_BLUE_PURPLE_8,
        'descr': 'light blue → purple',
    },
    'seq_blue_purple7': {
        'id': _vtk.vtkColorSeries.BREWER_SEQUENTIAL_BLUE_PURPLE_7,
        'descr': 'light blue → purple',
    },
    'seq_blue_purple6': {
        'id': _vtk.vtkColorSeries.BREWER_SEQUENTIAL_BLUE_PURPLE_6,
        'descr': 'light blue → purple',
    },
    'seq_blue_purple5': {
        'id': _vtk.vtkColorSeries.BREWER_SEQUENTIAL_BLUE_PURPLE_5,
        'descr': 'light blue → purple',
    },
    'seq_blue_purple4': {
        'id': _vtk.vtkColorSeries.BREWER_SEQUENTIAL_BLUE_PURPLE_4,
        'descr': 'light blue → purple',
    },
    'seq_blue_purple3': {
        'id': _vtk.vtkColorSeries.BREWER_SEQUENTIAL_BLUE_PURPLE_3,
        'descr': 'light blue → purple',
    },
    'qual_accent': {
        'id': _vtk.vtkColorSeries.BREWER_QUALITATIVE_ACCENT,
        'descr': 'pastel green, pastel purple, pastel orange, pastel yellow, blue, pink, '
        'brown, gray',
    },
    'qual_dark2': {
        'id': _vtk.vtkColorSeries.BREWER_QUALITATIVE_DARK2,
        'descr': 'darker shade of qual_set2',
    },
    'qual_set3': {
        'id': _vtk.vtkColorSeries.BREWER_QUALITATIVE_SET3,
        'descr': 'pastel colors: blue green, light yellow, dark purple, red, blue, orange, green, '
        'pink, gray, purple, light green, yellow',
    },
    'qual_set2': {
        'id': _vtk.vtkColorSeries.BREWER_QUALITATIVE_SET2,
        'descr': 'blue green, orange, purple, pink, green, yellow, brown, gray',
    },
    'qual_set1': {
        'id': _vtk.vtkColorSeries.BREWER_QUALITATIVE_SET1,
        'descr': 'red, blue, green, purple, orange, yellow, brown, pink, gray',
    },
    'qual_pastel2': {
        'id': _vtk.vtkColorSeries.BREWER_QUALITATIVE_PASTEL2,
        'descr': 'pastel shade of qual_set2',
    },
    'qual_pastel1': {
        'id': _vtk.vtkColorSeries.BREWER_QUALITATIVE_PASTEL1,
        'descr': 'pastel shade of qual_set1',
    },
    'qual_paired': {
        'id': _vtk.vtkColorSeries.BREWER_QUALITATIVE_PAIRED,
        'descr': 'light blue, blue, light green, green, light red, red, light orange, orange, '
        'light purple, purple, light yellow',
    },
    'custom': {'id': _vtk.vtkColorSeries.CUSTOM, 'descr': None},
}

SCHEME_NAMES = {
    scheme_info['id']: scheme_name  # type: ignore[index]
    for scheme_name, scheme_info in COLOR_SCHEMES.items()
}

# Define colormaps that require colorcet
# matches set(colorcet.cm.keys()) - set(mpl.colormaps)
_COLORCET_CMAPS_LITERAL = Literal[
    'CET_C1',
    'CET_C10',
    'CET_C10_r',
    'CET_C10s',
    'CET_C10s_r',
    'CET_C11',
    'CET_C11_r',
    'CET_C11s',
    'CET_C11s_r',
    'CET_C1_r',
    'CET_C1s',
    'CET_C1s_r',
    'CET_C2',
    'CET_C2_r',
    'CET_C2s',
    'CET_C2s_r',
    'CET_C3',
    'CET_C3_r',
    'CET_C3s',
    'CET_C3s_r',
    'CET_C4',
    'CET_C4_r',
    'CET_C4s',
    'CET_C4s_r',
    'CET_C5',
    'CET_C5_r',
    'CET_C5s',
    'CET_C5s_r',
    'CET_C6',
    'CET_C6_r',
    'CET_C6s',
    'CET_C6s_r',
    'CET_C7',
    'CET_C7_r',
    'CET_C7s',
    'CET_C7s_r',
    'CET_C8',
    'CET_C8_r',
    'CET_C8s',
    'CET_C8s_r',
    'CET_C9',
    'CET_C9_r',
    'CET_C9s',
    'CET_C9s_r',
    'CET_CBC1',
    'CET_CBC1_r',
    'CET_CBC2',
    'CET_CBC2_r',
    'CET_CBD1',
    'CET_CBD1_r',
    'CET_CBD2',
    'CET_CBD2_r',
    'CET_CBL1',
    'CET_CBL1_r',
    'CET_CBL2',
    'CET_CBL2_r',
    'CET_CBL3',
    'CET_CBL3_r',
    'CET_CBL4',
    'CET_CBL4_r',
    'CET_CBTC1',
    'CET_CBTC1_r',
    'CET_CBTC2',
    'CET_CBTC2_r',
    'CET_CBTD1',
    'CET_CBTD1_r',
    'CET_CBTL1',
    'CET_CBTL1_r',
    'CET_CBTL2',
    'CET_CBTL2_r',
    'CET_CBTL3',
    'CET_CBTL3_r',
    'CET_CBTL4',
    'CET_CBTL4_r',
    'CET_D1',
    'CET_D10',
    'CET_D10_r',
    'CET_D11',
    'CET_D11_r',
    'CET_D12',
    'CET_D12_r',
    'CET_D13',
    'CET_D13_r',
    'CET_D1A',
    'CET_D1A_r',
    'CET_D1_r',
    'CET_D2',
    'CET_D2_r',
    'CET_D3',
    'CET_D3_r',
    'CET_D4',
    'CET_D4_r',
    'CET_D6',
    'CET_D6_r',
    'CET_D7',
    'CET_D7_r',
    'CET_D8',
    'CET_D8_r',
    'CET_D9',
    'CET_D9_r',
    'CET_I1',
    'CET_I1_r',
    'CET_I2',
    'CET_I2_r',
    'CET_I3',
    'CET_I3_r',
    'CET_L1',
    'CET_L10',
    'CET_L10_r',
    'CET_L11',
    'CET_L11_r',
    'CET_L12',
    'CET_L12_r',
    'CET_L13',
    'CET_L13_r',
    'CET_L14',
    'CET_L14_r',
    'CET_L15',
    'CET_L15_r',
    'CET_L16',
    'CET_L16_r',
    'CET_L17',
    'CET_L17_r',
    'CET_L18',
    'CET_L18_r',
    'CET_L19',
    'CET_L19_r',
    'CET_L1_r',
    'CET_L2',
    'CET_L20',
    'CET_L20_r',
    'CET_L2_r',
    'CET_L3',
    'CET_L3_r',
    'CET_L4',
    'CET_L4_r',
    'CET_L5',
    'CET_L5_r',
    'CET_L6',
    'CET_L6_r',
    'CET_L7',
    'CET_L7_r',
    'CET_L8',
    'CET_L8_r',
    'CET_L9',
    'CET_L9_r',
    'CET_R1',
    'CET_R1_r',
    'CET_R2',
    'CET_R2_r',
    'CET_R3',
    'CET_R3_r',
    'CET_R4',
    'CET_R4_r',
    'bgy',
    'bgy_r',
    'bgyw',
    'bgyw_r',
    'bjy',
    'bjy_r',
    'bkr',
    'bkr_r',
    'bky',
    'bky_r',
    'blues',
    'blues_r',
    'bmw',
    'bmw_r',
    'bmy',
    'bmy_r',
    'bwy',
    'bwy_r',
    'circle_mgbm_67_c31',
    'circle_mgbm_67_c31_r',
    'circle_mgbm_67_c31_s25',
    'circle_mgbm_67_c31_s25_r',
    'colorwheel',
    'colorwheel_r',
    'cwr',
    'cwr_r',
    'cyclic_bgrmb_35_70_c75',
    'cyclic_bgrmb_35_70_c75_r',
    'cyclic_bgrmb_35_70_c75_s25',
    'cyclic_bgrmb_35_70_c75_s25_r',
    'cyclic_grey_15_85_c0',
    'cyclic_grey_15_85_c0_r',
    'cyclic_grey_15_85_c0_s25',
    'cyclic_grey_15_85_c0_s25_r',
    'cyclic_isoluminant',
    'cyclic_isoluminant_r',
    'cyclic_mrybm_35_75_c68',
    'cyclic_mrybm_35_75_c68_r',
    'cyclic_mrybm_35_75_c68_s25',
    'cyclic_mrybm_35_75_c68_s25_r',
    'cyclic_mybm_20_100_c48',
    'cyclic_mybm_20_100_c48_r',
    'cyclic_mybm_20_100_c48_s25',
    'cyclic_mybm_20_100_c48_s25_r',
    'cyclic_mygbm_30_95_c78',
    'cyclic_mygbm_30_95_c78_r',
    'cyclic_mygbm_30_95_c78_s25',
    'cyclic_mygbm_30_95_c78_s25_r',
    'cyclic_mygbm_50_90_c46',
    'cyclic_mygbm_50_90_c46_r',
    'cyclic_mygbm_50_90_c46_s25',
    'cyclic_mygbm_50_90_c46_s25_r',
    'cyclic_protanopic_deuteranopic_bwyk_16_96_c31',
    'cyclic_protanopic_deuteranopic_bwyk_16_96_c31_r',
    'cyclic_protanopic_deuteranopic_wywb_55_96_c33',
    'cyclic_protanopic_deuteranopic_wywb_55_96_c33_r',
    'cyclic_rygcbmr_50_90_c64',
    'cyclic_rygcbmr_50_90_c64_r',
    'cyclic_rygcbmr_50_90_c64_s25',
    'cyclic_rygcbmr_50_90_c64_s25_r',
    'cyclic_tritanopic_cwrk_40_100_c20',
    'cyclic_tritanopic_cwrk_40_100_c20_r',
    'cyclic_tritanopic_wrwc_70_100_c20',
    'cyclic_tritanopic_wrwc_70_100_c20_r',
    'cyclic_wrkbw_10_90_c43',
    'cyclic_wrkbw_10_90_c43_r',
    'cyclic_wrkbw_10_90_c43_s25',
    'cyclic_wrkbw_10_90_c43_s25_r',
    'cyclic_wrwbw_40_90_c42',
    'cyclic_wrwbw_40_90_c42_r',
    'cyclic_wrwbw_40_90_c42_s25',
    'cyclic_wrwbw_40_90_c42_s25_r',
    'cyclic_ymcgy_60_90_c67',
    'cyclic_ymcgy_60_90_c67_r',
    'cyclic_ymcgy_60_90_c67_s25',
    'cyclic_ymcgy_60_90_c67_s25_r',
    'dimgray',
    'dimgray_r',
    'diverging_bkr_55_10_c35',
    'diverging_bkr_55_10_c35_r',
    'diverging_bky_60_10_c30',
    'diverging_bky_60_10_c30_r',
    'diverging_bwg_20_95_c41',
    'diverging_bwg_20_95_c41_r',
    'diverging_bwr_20_95_c54',
    'diverging_bwr_20_95_c54_r',
    'diverging_bwr_40_95_c42',
    'diverging_bwr_40_95_c42_r',
    'diverging_bwr_55_98_c37',
    'diverging_bwr_55_98_c37_r',
    'diverging_cwm_80_100_c22',
    'diverging_cwm_80_100_c22_r',
    'diverging_gkr_60_10_c40',
    'diverging_gkr_60_10_c40_r',
    'diverging_gwr_55_95_c38',
    'diverging_gwr_55_95_c38_r',
    'diverging_gwv_55_95_c39',
    'diverging_gwv_55_95_c39_r',
    'diverging_isoluminant_cjm_75_c23',
    'diverging_isoluminant_cjm_75_c23_r',
    'diverging_isoluminant_cjm_75_c24',
    'diverging_isoluminant_cjm_75_c24_r',
    'diverging_isoluminant_cjo_70_c25',
    'diverging_isoluminant_cjo_70_c25_r',
    'diverging_linear_bjr_30_55_c53',
    'diverging_linear_bjr_30_55_c53_r',
    'diverging_linear_bjy_30_90_c45',
    'diverging_linear_bjy_30_90_c45_r',
    'diverging_linear_protanopic_deuteranopic_bjy_57_89_c34',
    'diverging_linear_protanopic_deuteranopic_bjy_57_89_c34_r',
    'diverging_protanopic_deuteranopic_bwy_60_95_c32',
    'diverging_protanopic_deuteranopic_bwy_60_95_c32_r',
    'diverging_rainbow_bgymr_45_85_c67',
    'diverging_rainbow_bgymr_45_85_c67_r',
    'diverging_tritanopic_cwr_75_98_c20',
    'diverging_tritanopic_cwr_75_98_c20_r',
    'fire',
    'fire_r',
    'glasbey',
    'glasbey_bw',
    'glasbey_bw_minc_20',
    'glasbey_bw_minc_20_hue_150_280',
    'glasbey_bw_minc_20_hue_150_280_r',
    'glasbey_bw_minc_20_hue_330_100',
    'glasbey_bw_minc_20_hue_330_100_r',
    'glasbey_bw_minc_20_maxl_70',
    'glasbey_bw_minc_20_maxl_70_r',
    'glasbey_bw_minc_20_minl_30',
    'glasbey_bw_minc_20_minl_30_r',
    'glasbey_bw_minc_20_r',
    'glasbey_bw_r',
    'glasbey_category10',
    'glasbey_category10_r',
    'glasbey_cool',
    'glasbey_cool_r',
    'glasbey_dark',
    'glasbey_dark_r',
    'glasbey_hv',
    'glasbey_hv_r',
    'glasbey_light',
    'glasbey_light_r',
    'glasbey_r',
    'glasbey_warm',
    'glasbey_warm_r',
    'gouldian',
    'gouldian_r',
    'gwv',
    'gwv_r',
    'isolum',
    'isolum_r',
    'isoluminant_cgo_70_c39',
    'isoluminant_cgo_70_c39_r',
    'isoluminant_cgo_80_c38',
    'isoluminant_cgo_80_c38_r',
    'isoluminant_cm_70_c39',
    'isoluminant_cm_70_c39_r',
    'kb',
    'kb_r',
    'kbc',
    'kbc_r',
    'kbgyw',
    'kbgyw_r',
    'kg',
    'kg_r',
    'kgy',
    'kgy_r',
    'kr',
    'kr_r',
    'linear_bgy_10_95_c74',
    'linear_bgy_10_95_c74_r',
    'linear_bgyw_15_100_c67',
    'linear_bgyw_15_100_c67_r',
    'linear_bgyw_15_100_c68',
    'linear_bgyw_15_100_c68_r',
    'linear_bgyw_20_98_c66',
    'linear_bgyw_20_98_c66_r',
    'linear_blue_5_95_c73',
    'linear_blue_5_95_c73_r',
    'linear_blue_95_50_c20',
    'linear_blue_95_50_c20_r',
    'linear_bmw_5_95_c86',
    'linear_bmw_5_95_c86_r',
    'linear_bmw_5_95_c89',
    'linear_bmw_5_95_c89_r',
    'linear_bmy_10_95_c71',
    'linear_bmy_10_95_c71_r',
    'linear_bmy_10_95_c78',
    'linear_bmy_10_95_c78_r',
    'linear_gow_60_85_c27',
    'linear_gow_60_85_c27_r',
    'linear_gow_65_90_c35',
    'linear_gow_65_90_c35_r',
    'linear_green_5_95_c69',
    'linear_green_5_95_c69_r',
    'linear_grey_0_100_c0',
    'linear_grey_0_100_c0_r',
    'linear_grey_10_95_c0',
    'linear_grey_10_95_c0_r',
    'linear_kbc_5_95_c73',
    'linear_kbc_5_95_c73_r',
    'linear_kbgoy_20_95_c57',
    'linear_kbgoy_20_95_c57_r',
    'linear_kbgyw_10_98_c63',
    'linear_kbgyw_10_98_c63_r',
    'linear_kbgyw_5_98_c62',
    'linear_kbgyw_5_98_c62_r',
    'linear_kgy_5_95_c69',
    'linear_kgy_5_95_c69_r',
    'linear_kry_0_97_c73',
    'linear_kry_0_97_c73_r',
    'linear_kry_5_95_c72',
    'linear_kry_5_95_c72_r',
    'linear_kry_5_98_c75',
    'linear_kry_5_98_c75_r',
    'linear_kryw_0_100_c71',
    'linear_kryw_0_100_c71_r',
    'linear_kryw_5_100_c64',
    'linear_kryw_5_100_c64_r',
    'linear_kryw_5_100_c67',
    'linear_kryw_5_100_c67_r',
    'linear_protanopic_deuteranopic_kbjyw_5_95_c25',
    'linear_protanopic_deuteranopic_kbjyw_5_95_c25_r',
    'linear_protanopic_deuteranopic_kbw_5_95_c34',
    'linear_protanopic_deuteranopic_kbw_5_95_c34_r',
    'linear_protanopic_deuteranopic_kbw_5_98_c40',
    'linear_protanopic_deuteranopic_kbw_5_98_c40_r',
    'linear_protanopic_deuteranopic_kyw_5_95_c49',
    'linear_protanopic_deuteranopic_kyw_5_95_c49_r',
    'linear_ternary_blue_0_44_c57',
    'linear_ternary_blue_0_44_c57_r',
    'linear_ternary_green_0_46_c42',
    'linear_ternary_green_0_46_c42_r',
    'linear_ternary_red_0_50_c52',
    'linear_ternary_red_0_50_c52_r',
    'linear_tritanopic_kcw_5_95_c22',
    'linear_tritanopic_kcw_5_95_c22_r',
    'linear_tritanopic_krjcw_5_95_c24',
    'linear_tritanopic_krjcw_5_95_c24_r',
    'linear_tritanopic_krjcw_5_98_c46',
    'linear_tritanopic_krjcw_5_98_c46_r',
    'linear_tritanopic_krw_5_95_c46',
    'linear_tritanopic_krw_5_95_c46_r',
    'linear_wcmr_100_45_c42',
    'linear_wcmr_100_45_c42_r',
    'linear_worb_100_25_c53',
    'linear_worb_100_25_c53_r',
    'linear_wyor_100_45_c55',
    'linear_wyor_100_45_c55_r',
    'rainbow4',
    'rainbow4_r',
    'rainbow_bgyr_10_90_c83',
    'rainbow_bgyr_10_90_c83_r',
    'rainbow_bgyr_35_85_c72',
    'rainbow_bgyr_35_85_c72_r',
    'rainbow_bgyr_35_85_c73',
    'rainbow_bgyr_35_85_c73_r',
    'rainbow_bgyrm_35_85_c69',
    'rainbow_bgyrm_35_85_c69_r',
    'rainbow_bgyrm_35_85_c71',
    'rainbow_bgyrm_35_85_c71_r',
]
_COLORCET_CMAPS = get_args(_COLORCET_CMAPS_LITERAL)

# Define colormaps that require cmocean
# matches set(cmocean.cm.cmap_d.keys()) - set(mpl.colormaps)
_CMOCEAN_CMAPS_LITERAL = Literal[
    'algae',
    'algae_i',
    'algae_i_r',
    'algae_r',
    'algae_r_i',
    'amp',
    'amp_i',
    'amp_i_r',
    'amp_r',
    'amp_r_i',
    'balance',
    'balance_i',
    'balance_i_r',
    'balance_r',
    'balance_r_i',
    'curl',
    'curl_i',
    'curl_i_r',
    'curl_r',
    'curl_r_i',
    'deep',
    'deep_i',
    'deep_i_r',
    'deep_r',
    'deep_r_i',
    'delta',
    'delta_i',
    'delta_i_r',
    'delta_r',
    'delta_r_i',
    'dense',
    'dense_i',
    'dense_i_r',
    'dense_r',
    'dense_r_i',
    'diff',
    'diff_i',
    'diff_i_r',
    'diff_r',
    'diff_r_i',
    'gray_i',
    'gray_i_r',
    'gray_r_i',
    'haline',
    'haline_i',
    'haline_i_r',
    'haline_r',
    'haline_r_i',
    'ice',
    'ice_i',
    'ice_i_r',
    'ice_r',
    'ice_r_i',
    'matter',
    'matter_i',
    'matter_i_r',
    'matter_r',
    'matter_r_i',
    'oxy',
    'oxy_i',
    'oxy_i_r',
    'oxy_r',
    'oxy_r_i',
    'phase',
    'phase_i',
    'phase_i_r',
    'phase_r',
    'phase_r_i',
    'rain',
    'rain_i',
    'rain_i_r',
    'rain_r',
    'rain_r_i',
    'solar',
    'solar_i',
    'solar_i_r',
    'solar_r',
    'solar_r_i',
    'speed',
    'speed_i',
    'speed_i_r',
    'speed_r',
    'speed_r_i',
    'tarn',
    'tarn_i',
    'tarn_i_r',
    'tarn_r',
    'tarn_r_i',
    'tempo',
    'tempo_i',
    'tempo_i_r',
    'tempo_r',
    'tempo_r_i',
    'thermal',
    'thermal_i',
    'thermal_i_r',
    'thermal_r',
    'thermal_r_i',
    'topo',
    'topo_i',
    'topo_i_r',
    'topo_r',
    'topo_r_i',
    'turbid',
    'turbid_i',
    'turbid_i_r',
    'turbid_r',
    'turbid_r_i',
]
_CMOCEAN_CMAPS = get_args(_CMOCEAN_CMAPS_LITERAL)

_CMCRAMERI_CMAPS_LITERAL = Literal[
    'acton',
    'actonS',
    'acton_r',
    'bam',
    'bamO',
    'bamO_r',
    'bam_r',
    'bamako',
    'bamakoS',
    'bamako_r',
    'batlow',
    'batlowK',
    'batlowKS',
    'batlowK_r',
    'batlowS',
    'batlowW',
    'batlowWS',
    'batlowW_r',
    'batlow_r',
    'bilbao',
    'bilbaoS',
    'bilbao_r',
    'broc',
    'brocO',
    'brocO_r',
    'broc_r',
    'buda',
    'budaS',
    'buda_r',
    'bukavu',
    'bukavu_r',
    'cork',
    'corkO',
    'corkO_r',
    'cork_r',
    'davos',
    'davosS',
    'davos_r',
    'devon',
    'devonS',
    'devon_r',
    'fes',
    'fes_r',
    'glasgow',
    'glasgowS',
    'glasgow_r',
    'grayC',
    'grayCS',
    'grayC_r',
    'hawaii',
    'hawaiiS',
    'hawaii_r',
    'imola',
    'imolaS',
    'imola_r',
    'lajolla',
    'lajollaS',
    'lajolla_r',
    'lapaz',
    'lapazS',
    'lapaz_r',
    'lipari',
    'lipariS',
    'lipari_r',
    'lisbon',
    'lisbon_r',
    'navia',
    'naviaS',
    'navia_r',
    'nuuk',
    'nuukS',
    'nuuk_r',
    'oleron',
    'oleron_r',
    'oslo',
    'osloS',
    'oslo_r',
    'roma',
    'romaO',
    'romaO_r',
    'roma_r',
    'tofino',
    'tofino_r',
    'tokyo',
    'tokyoS',
    'tokyo_r',
    'turku',
    'turkuS',
    'turku_r',
    'vik',
    'vikO',
    'vikO_r',
    'vik_r',
]
_CMCRAMERI_CMAPS = get_args(_CMCRAMERI_CMAPS_LITERAL)

_MATPLOTLIB_CMAPS_LITERAL = Literal[
    'Accent',
    'Accent_r',
    'Blues',
    'Blues_r',
    'BrBG',
    'BrBG_r',
    'BuGn',
    'BuGn_r',
    'BuPu',
    'BuPu_r',
    'CMRmap',
    'CMRmap_r',
    'Dark2',
    'Dark2_r',
    'GnBu',
    'GnBu_r',
    'Grays',
    'Grays_r',
    'Greens',
    'Greens_r',
    'Greys',
    'Greys_r',
    'OrRd',
    'OrRd_r',
    'Oranges',
    'Oranges_r',
    'PRGn',
    'PRGn_r',
    'Paired',
    'Paired_r',
    'Pastel1',
    'Pastel1_r',
    'Pastel2',
    'Pastel2_r',
    'PiYG',
    'PiYG_r',
    'PuBu',
    'PuBuGn',
    'PuBuGn_r',
    'PuBu_r',
    'PuOr',
    'PuOr_r',
    'PuRd',
    'PuRd_r',
    'Purples',
    'Purples_r',
    'RdBu',
    'RdBu_r',
    'RdGy',
    'RdGy_r',
    'RdPu',
    'RdPu_r',
    'RdYlBu',
    'RdYlBu_r',
    'RdYlGn',
    'RdYlGn_r',
    'Reds',
    'Reds_r',
    'Set1',
    'Set1_r',
    'Set2',
    'Set2_r',
    'Set3',
    'Set3_r',
    'Spectral',
    'Spectral_r',
    'Wistia',
    'Wistia_r',
    'YlGn',
    'YlGnBu',
    'YlGnBu_r',
    'YlGn_r',
    'YlOrBr',
    'YlOrBr_r',
    'YlOrRd',
    'YlOrRd_r',
    'afmhot',
    'afmhot_r',
    'autumn',
    'autumn_r',
    'berlin',
    'berlin_r',
    'binary',
    'binary_r',
    'bone',
    'bone_r',
    'brg',
    'brg_r',
    'bwr',
    'bwr_r',
    'cividis',
    'cividis_r',
    'cool',
    'cool_r',
    'coolwarm',
    'coolwarm_r',
    'copper',
    'copper_r',
    'cubehelix',
    'cubehelix_r',
    'flag',
    'flag_r',
    'gist_earth',
    'gist_earth_r',
    'gist_gray',
    'gist_gray_r',
    'gist_grey',
    'gist_grey_r',
    'gist_heat',
    'gist_heat_r',
    'gist_ncar',
    'gist_ncar_r',
    'gist_rainbow',
    'gist_rainbow_r',
    'gist_stern',
    'gist_stern_r',
    'gist_yarg',
    'gist_yarg_r',
    'gist_yerg',
    'gist_yerg_r',
    'gnuplot',
    'gnuplot2',
    'gnuplot2_r',
    'gnuplot_r',
    'gray',
    'gray_r',
    'grey',
    'grey_r',
    'hot',
    'hot_r',
    'hsv',
    'hsv_r',
    'inferno',
    'inferno_r',
    'jet',
    'jet_r',
    'magma',
    'magma_r',
    'managua',
    'managua_r',
    'nipy_spectral',
    'nipy_spectral_r',
    'ocean',
    'ocean_r',
    'okabe_ito',
    'okabe_ito_r',
    'pink',
    'pink_r',
    'plasma',
    'plasma_r',
    'prism',
    'prism_r',
    'rainbow',
    'rainbow_r',
    'seismic',
    'seismic_r',
    'spring',
    'spring_r',
    'summer',
    'summer_r',
    'tab10',
    'tab10_r',
    'tab20',
    'tab20_r',
    'tab20b',
    'tab20b_r',
    'tab20c',
    'tab20c_r',
    'terrain',
    'terrain_r',
    'turbo',
    'turbo_r',
    'twilight',
    'twilight_r',
    'twilight_shifted',
    'twilight_shifted_r',
    'vanimo',
    'vanimo_r',
    'viridis',
    'viridis_r',
    'winter',
    'winter_r',
]

_MATPLOTLIB_CMAPS = get_args(_MATPLOTLIB_CMAPS_LITERAL)


class Color(_NoNewAttrMixin):
    """Helper class to convert between different color representations used in the pyvista library.

    Many pyvista methods accept :data:`ColorLike` parameters. This helper class
    is used to convert such parameters to the necessary format, used by
    underlying (VTK) methods. Any color name (``str``), hex string (``str``)
    or RGB(A) sequence (``tuple``, ``list`` or ``numpy.ndarray`` of ``int``
    or ``float``) is considered a :data:`ColorLike` parameter and can be converted
    by this class.

    See :ref:`named_colors` for a list of supported colors.

    .. note:

        The internally used representation is an integer RGBA sequence (values
        between 0 and 255). This might however change in future releases.

    Parameters
    ----------
    color : ColorLike, optional
        Either a string, RGB sequence, RGBA sequence, or hex color string.
        RGB(A) sequences should either be provided as floats between 0 and 1
        or as ints between 0 and 255. Hex color strings can contain optional
        ``'#'`` or ``'0x'`` prefixes. If no opacity is provided, the
        ``default_opacity`` will be used. If ``color`` is ``None``, the
        ``default_color`` is used instead.
        The following examples all denote the color 'white':

        * ``'white'``
        * ``'w'``
        * ``[1.0, 1.0, 1.0]``
        * ``[255, 255, 255, 255]``
        * ``'#FFFFFF'``

    opacity : int | float | str, optional
        Opacity of the represented color. Overrides any opacity associated
        with the provided ``color``. Allowed opacities are floats between 0
        and 1, ints between 0 and 255 or hexadecimal strings of length 2
        (plus the length of the optional prefix).
        The following examples all denote a fully opaque color:

        * ``1.0``
        * ``255``
        * ``'#ff'``

    default_color : ColorLike, optional
        Default color to use when ``color`` is ``None``. If this value is
        ``None``, then defaults to the global theme color. Format is
        identical to ``color``.

    default_opacity : int | float | str, optional
        Default opacity of the represented color. Used when ``color``
        does not specify an opacity and ``opacity`` is ``None``. Format
        is identical to ``opacity``.

    Examples
    --------
    Create a transparent green color using a color name, float RGBA sequence,
    integer RGBA sequence and RGBA hexadecimal string.

    >>> import pyvista as pv
    >>> pv.Color('green', opacity=0.5)
    Color(name='green', hex='#00800080', opacity=128)
    >>> pv.Color([0.0, 0.5, 0.0, 0.5])
    Color(name='green', hex='#00800080', opacity=128)
    >>> pv.Color([0, 128, 0, 128])
    Color(name='green', hex='#00800080', opacity=128)
    >>> pv.Color('#00800080')
    Color(name='green', hex='#00800080', opacity=128)

    """

    # Supported names for each color channel.
    CHANNEL_NAMES = (
        {'red', 'r'},  # 0
        {'green', 'g'},  # 1
        {'blue', 'b'},  # 2
        {'alpha', 'a', 'opacity'},  # 3
    )

    @_deprecate_positional_args(allowed=['color', 'opacity'])
    def __init__(  # noqa: PLR0917
        self,
        color: ColorLike | None = None,
        opacity: float | str | None = None,
        default_color: ColorLike | None = None,
        default_opacity: float | str = 255,
    ):
        """Initialize new instance."""
        self._red, self._green, self._blue, self._opacity = 0, 0, 0, 0
        self._opacity = self.convert_color_channel(default_opacity)
        self._name = None

        # Use default color if no color is provided
        if color is None:
            color = pv.global_theme.color if default_color is None else default_color

        _validation.check_instance(
            color, (Color, str, dict, list, tuple, np.ndarray, _vtk.vtkColor3ub), name='color'
        )
        try:
            if isinstance(color, Color):
                # Create copy of color instance
                self._red, self._green, self._blue, self._opacity = color.int_rgba
            elif isinstance(color, str):
                # From named color or hex string
                self._from_str(color)
            elif isinstance(color, dict):
                # From dictionary
                self._from_dict(color)
            elif isinstance(color, (list, tuple, np.ndarray)):
                # From RGB(A) sequence
                self._from_rgba(color)
            elif isinstance(color, _vtk.vtkColor3ub):
                # From vtkColor3ub instance (can be unpacked as rgb tuple)
                self._from_rgba(color)
            else:  # pragma: no cover
                msg = f'Unexpected color type: {type(color)}'
                raise TypeError(msg)
            self._name = color_names.get(self.hex_rgb)
        except ValueError as e:
            msg = (
                '\n'
                f'\tInvalid color input: ({color})\n'
                '\tMust be a string, rgb(a) sequence, or hex color string.  For example:\n'
                "\t\tcolor='white'\n"
                "\t\tcolor='w'\n"
                '\t\tcolor=[1.0, 1.0, 1.0]\n'
                '\t\tcolor=[255, 255, 255]\n'
                "\t\tcolor='#FFFFFF'"
            )
            raise ValueError(msg) from e

        # Overwrite opacity if it is provided
        try:
            if opacity is not None:
                self._opacity = self.convert_color_channel(opacity)
        except ValueError as e:
            msg = (
                '\n'
                f'\tInvalid opacity input: ({opacity})'
                '\tMust be an integer, float or string.  For example:\n'
                "\t\topacity='1.0'\n"
                "\t\topacity='255'\n"
                "\t\topacity='#FF'"
            )
            raise ValueError(msg) from e

    @staticmethod
    def strip_hex_prefix(h: str) -> str:
        """Strip any ``'#'`` or ``'0x'`` prefix from a hexadecimal string.

        Parameters
        ----------
        h : str
            Hexadecimal string to strip.

        Returns
        -------
        str
            Stripped hexadecimal string.

        """
        h = h.lstrip('#')
        return h.removeprefix('0x')

    @staticmethod
    def convert_color_channel(
        val: float | np.floating[Any] | np.integer[Any] | str,
    ) -> int:
        """Convert the given color channel value to the integer representation.

        Parameters
        ----------
        val : int | float | str
            Color channel value to convert. Supported input values are a
            hex string of length 2 (``'00'`` to ``'ff'``) with an optional
            prefix (``'#'`` or ``'0x'``), a float (``0.0`` to ``1.0``) or
            an integer (``0`` to ``255``).

        Returns
        -------
        int
            Color channel value in the integer representation (values between
            ``0`` and ``255``).

        """
        # Check for numpy inputs to avoid unnecessary calls to np.issubdtype
        arr = None
        if isinstance(val, (np.ndarray, np.generic)):
            arr = np.asanyarray(val)

        # Convert non-integers to int
        if isinstance(val, str):
            # From hexadecimal value
            val = int(Color.strip_hex_prefix(val), 16)
        elif isinstance(val, float) or (
            arr is not None and np.issubdtype(arr.dtype, np.floating) and arr.ndim == 0
        ):
            val = round(255 * val)

        # Check integers
        if isinstance(val, int) and 0 <= val <= 255:
            return val  # type: ignore[return-value]
        elif isinstance(val, np.uint8) or (
            arr is not None
            and np.issubdtype(arr.dtype, np.integer)
            and arr.ndim == 0
            and 0 <= val <= 255
        ):
            return int(val)
        else:
            msg = f'Unsupported color channel value provided: {val}'
            raise ValueError(msg)

    def _from_rgba(self, rgba):
        """Construct color from an RGB(A) sequence."""
        arg = rgba
        if len(rgba) == 3:
            # Keep using current opacity if it is not provided.
            rgba = [*rgba, self._opacity]
        try:
            if len(rgba) != 4:
                msg = 'Invalid length for RGBA sequence.'
                raise ValueError(msg)
            self._red, self._green, self._blue, self._opacity = (
                self.convert_color_channel(c) for c in rgba
            )
        except ValueError:
            msg = f'Invalid RGB(A) sequence: {arg}'
            raise ValueError(msg) from None

    def _from_dict(self, dct):
        """Construct color from an RGB(A) dictionary."""
        # Get any of the keys associated with each color channel (or None).
        rgba = [
            next((dct[key] for key in cnames if key in dct), None) for cnames in self.CHANNEL_NAMES
        ]
        self._from_rgba(rgba)

    def _from_hex(self, h):
        """Construct color from a hex string."""
        arg = h
        h = self.strip_hex_prefix(h)
        try:
            self._from_rgba(
                [self.convert_color_channel(h[i : i + 2]) for i in range(0, len(h), 2)]
            )
        except ValueError:
            msg = f'Invalid hex string: {arg}'
            raise ValueError(msg) from None

    def _from_str(self, n: str):
        """Construct color from a name or hex string."""
        arg = n
        n = _format_color_name(n)
        if n in color_synonyms:
            # Synonym of registered color name
            # Convert from synonym to full hex
            n = color_synonyms[n]
            self._from_hex(hexcolors[n])
        elif n in hexcolors:
            # Color name
            self._from_hex(hexcolors[n])
        else:
            # Otherwise, try conversion to hex
            try:
                self._from_hex(n)
            except ValueError:
                msg = f'Invalid color name or hex string: {arg}'
                raise ValueError(msg) from None

    @property
    def int_rgba(self) -> tuple[int, int, int, int]:  # numpydoc ignore=RT01
        """Get the color value as an RGBA integer tuple.

        Examples
        --------
        Create a blue color with half opacity.

        >>> import pyvista as pv
        >>> c = pv.Color('blue', opacity=128)
        >>> c
        Color(name='blue', hex='#0000ff80', opacity=128)
        >>> c.int_rgba
        (0, 0, 255, 128)

        Create a transparent red color using an integer RGBA sequence.

        >>> c = pv.Color([255, 0, 0, 64])
        >>> c
        Color(name='red', hex='#ff000040', opacity=64)
        >>> c.int_rgba
        (255, 0, 0, 64)

        """
        return self._red, self._green, self._blue, self._opacity

    @property
    def int_rgb(self) -> tuple[int, int, int]:  # numpydoc ignore=RT01
        """Get the color value as an RGB integer tuple.

        Examples
        --------
        Create a blue color with half opacity.

        >>> import pyvista as pv
        >>> c = pv.Color('blue', opacity=128)
        >>> c
        Color(name='blue', hex='#0000ff80', opacity=128)
        >>> c.int_rgb
        (0, 0, 255)

        Create a red color using an integer RGB sequence.

        >>> c = pv.Color([255, 0, 0])
        >>> c
        Color(name='red', hex='#ff0000ff', opacity=255)
        >>> c.int_rgb
        (255, 0, 0)

        """
        return self.int_rgba[:3]

    @property
    def float_rgba(self) -> tuple[float, float, float, float]:  # numpydoc ignore=RT01
        """Get the color value as an RGBA float tuple.

        Examples
        --------
        Create a blue color with custom opacity.

        >>> import pyvista as pv
        >>> c = pv.Color('blue', opacity=0.6)
        >>> c
        Color(name='blue', hex='#0000ff99', opacity=153)
        >>> c.float_rgba
        (0.0, 0.0, 1.0, 0.6)

        Create a transparent red color using a float RGBA sequence.

        >>> c = pv.Color([1.0, 0.0, 0.0, 0.2])
        >>> c
        Color(name='red', hex='#ff000033', opacity=51)
        >>> c.float_rgba
        (1.0, 0.0, 0.0, 0.2)

        """
        return (
            self._red / 255.0,
            self._green / 255.0,
            self._blue / 255.0,
            self._opacity / 255.0,
        )

    @property
    def float_rgb(self) -> tuple[float, float, float]:  # numpydoc ignore=RT01
        """Get the color value as an RGB float tuple.

        Examples
        --------
        Create a blue color with custom opacity.

        >>> import pyvista as pv
        >>> c = pv.Color('blue', default_opacity=0.6)
        >>> c
        Color(name='blue', hex='#0000ff99', opacity=153)
        >>> c.float_rgb
        (0.0, 0.0, 1.0)

        Create a red color using a float RGB sequence.

        >>> c = pv.Color([1.0, 0.0, 0.0])
        >>> c
        Color(name='red', hex='#ff0000ff', opacity=255)
        >>> c.float_rgb
        (1.0, 0.0, 0.0)

        """
        return self.float_rgba[:3]

    @property
    def _float_hls(self) -> tuple[float, float, float]:
        """Get the color as Hue, Lightness, Saturation (HLS) in range [0.0, 1.0]."""
        return rgb_to_hls(*self.float_rgb)

    @property
    def hex_rgba(self) -> str:  # numpydoc ignore=RT01
        """Get the color value as an RGBA hexadecimal value.

        Examples
        --------
        Create a blue color with half opacity.

        >>> import pyvista as pv
        >>> c = pv.Color('blue', default_opacity='#80')
        >>> c
        Color(name='blue', hex='#0000ff80', opacity=128)
        >>> c.hex_rgba
        '#0000ff80'

        Create a transparent red color using an RGBA hexadecimal value.

        >>> c = pv.Color('0xff000040')
        >>> c
        Color(name='red', hex='#ff000040', opacity=64)
        >>> c.hex_rgba
        '#ff000040'

        """
        return '#' + ''.join(
            f'{c:0>2x}' for c in (self._red, self._green, self._blue, self._opacity)
        )

    @property
    def hex_rgb(self) -> str:  # numpydoc ignore=RT01
        """Get the color value as an RGB hexadecimal value.

        Examples
        --------
        Create a blue color with half opacity.

        >>> import pyvista as pv
        >>> c = pv.Color('blue', default_opacity='#80')
        >>> c
        Color(name='blue', hex='#0000ff80', opacity=128)
        >>> c.hex_rgb
        '#0000ff'

        Create a red color using an RGB hexadecimal value.

        >>> c = pv.Color('0xff0000')
        >>> c
        Color(name='red', hex='#ff0000ff', opacity=255)
        >>> c.hex_rgb
        '#ff0000'

        """
        return self.hex_rgba[:-2]

    @property
    def name(self) -> str | None:  # numpydoc ignore=RT01
        """Get the color name.

        The name is always formatted as a lower case string without
        any delimiters.

        See :ref:`named_colors` for a list of supported colors.

        Returns
        -------
        output : str | None
            The color name, in case this color has a name; otherwise ``None``.

        Examples
        --------
        Create a dark blue color with half opacity.

        >>> import pyvista as pv
        >>> c = pv.Color('darkblue', default_opacity=0.5)
        >>> c
        Color(name='darkblue', hex='#00008b80', opacity=128)

        When creating a new ``Color``, the name may be delimited with a space,
        hyphen, underscore, or written as a single word.

        >>> c = pv.Color('dark blue', default_opacity=0.5)

        Upper-case letters are also supported.

        >>> c = pv.Color('DarkBlue', default_opacity=0.5)

        However, the name is always standardized as a single lower-case word.

        >>> c
        Color(name='darkblue', hex='#00008b80', opacity=128)

        """
        return self._name

    @property
    def vtk_c3ub(self) -> _vtk.vtkColor3ub:  # numpydoc ignore=RT01
        """Get the color value as a VTK Color3ub instance.

        Examples
        --------
        Create a blue color with half opacity.

        >>> import pyvista as pv
        >>> c = pv.Color('blue', default_opacity=0.5)
        >>> c
        Color(name='blue', hex='#0000ff80', opacity=128)
        >>> c.vtk_c3ub
        vtkmodules.vtkCommonDataModel.vtkColor3ub([0, 0, 255])

        """
        return _vtk.vtkColor3ub(self._red, self._green, self._blue)

    def linear_to_srgb(self):
        """Convert from linear color values to sRGB color values.

        Returns
        -------
        Color
            A new ``Color`` instance with sRGB color values.

        """
        rgba = np.array(self.float_rgba)
        mask = rgba < 0.0031308
        rgba[mask] *= 12.92
        rgba[~mask] = 1.055 * rgba[~mask] ** (1 / 2.4) - 0.055
        return Color(rgba)

    def srgb_to_linear(self):
        """Convert from sRGB color values to linear color values.

        Returns
        -------
        Color
            A new ``Color`` instance with linear color values.

        """
        rgba = np.array(self.float_rgba)
        mask = rgba < 0.04045
        rgba[mask] /= 12.92
        rgba[~mask] = ((rgba[~mask] + 0.055) / 1.055) ** 2.4
        return Color(rgba)

    @classmethod
    def from_dict(cls, dict_):  # numpydoc ignore=RT01
        """Construct from dictionary for JSON deserialization."""
        return Color(dict_)

    def to_dict(self):  # numpydoc ignore=RT01
        """Convert to dictionary for JSON serialization."""
        return {'r': self._red, 'g': self._green, 'b': self._blue, 'a': self._opacity}

    @property
    def opacity(self):  # numpydoc ignore=RT01
        """Return the opacity of this color in the range of ``(0-255)``.

        Examples
        --------
        >>> import pyvista as pv
        >>> color = pv.Color('r', opacity=0.5)
        >>> color.opacity
        128
        >>> color
        Color(name='red', hex='#ff000080', opacity=128)

        """
        return self._opacity

    def __eq__(self, other):
        """Equality comparison."""
        try:
            return self.int_rgba == Color(other).int_rgba
        except ValueError:  # pragma: no cover
            return NotImplemented

    def __hash__(self):  # pragma: no cover
        """Hash calculation."""
        return hash((self._red, self._green, self._blue, self._opacity))

    def __getitem__(self, item):
        """Support indexing the float RGBA representation for backward compatibility."""
        if not isinstance(item, (str, slice, int, np.integer)):
            msg = 'Invalid index specified, only strings and integers are supported.'
            raise TypeError(msg)
        if isinstance(item, str):
            for i, cnames in enumerate(self.CHANNEL_NAMES):
                if item in cnames:
                    item = i
                    break
            else:
                msg = f'Invalid string index {item!r}.'
                raise ValueError(msg)
        return self.float_rgba[item]

    def __iter__(self):
        """Support iteration over the float RGBA representation for backward compatibility."""
        return iter(self.float_rgba)

    def __repr__(self) -> str:  # pragma: no cover
        """Human readable representation."""
        kwargs = f'hex={self.hex_rgba!r}, opacity={self.opacity}'
        if self._name is not None:
            kwargs = f'name={self._name!r}, ' + kwargs
        return f'Color({kwargs})'


PARAVIEW_BACKGROUND = Color('paraview').float_rgb  # [82, 87, 110] / 255


def get_cmap_safe(cmap: ColormapOptions) -> colors.Colormap:
    """Fetch a colormap by name from matplotlib, colorcet, cmocean, or cmcrameri.

    See :ref:`named_colormaps` for supported colormaps.

    Parameters
    ----------
    cmap : str | list[str] | matplotlib.colors.Colormap
        Name of the colormap to fetch. If the input is a list of strings, the
        strings must be color names (from :ref:`named_colors`), and
        it will create a ``ListedColormap`` with the input list.

    Returns
    -------
    matplotlib.colors.Colormap
        The requested colormap if available.

    Raises
    ------
    ValueError
        If the input colormap name is not valid.
    TypeError
        If the input is a list of items that are not strings.

    """
    _validation.check_instance(cmap, (str, list, colors.Colormap), name='cmap')

    def get_3rd_party_cmap(cmap_):
        cmap_sources = {
            'colorcet.cm': _COLORCET_CMAPS,
            'cmocean.cm.cmap_d': _CMOCEAN_CMAPS,
            'cmcrameri.cm.cmaps': _CMCRAMERI_CMAPS,
        }

        def get_nested_attr(obj, attr_path):
            for attr in attr_path:
                obj = getattr(obj, attr)
            return obj

        # Try importing and returning cmap from each package
        for cmap_import, known_cmaps in cmap_sources.items():
            parts = cmap_import.split('.')
            top_module = parts[0]

            with contextlib.suppress(ImportError):
                mod = importlib.import_module(top_module)
                cmap_dict = get_nested_attr(mod, parts[1:])
                with contextlib.suppress(KeyError):
                    return cmap_dict[cmap_]

            if cmap_ in known_cmaps:  # pragma: no cover
                msg = (
                    f'Package `{top_module}` is required to use colormap {cmap_!r}.\n'
                    'Install PyVista with `pyvista[colormaps]` to install it by default.'
                )
                raise ModuleNotFoundError(msg)
        return None

    if isinstance(cmap, colors.Colormap):
        return cmap
    if isinstance(cmap, str):
        # check if this colormap has been mapped between ipygany
        if cmap in IPYGANY_MAP:
            cmap = IPYGANY_MAP[cmap]  # type: ignore[assignment]

        cmap_3rd_party = get_3rd_party_cmap(cmap)
        if cmap_3rd_party:
            return cmap_3rd_party
        elif not isinstance(cmap, colors.Colormap):
            if inspect.ismodule(colormaps):  # pragma: no cover
                # Backwards compatibility with matplotlib<3.5.0
                if not hasattr(colormaps, cmap):
                    msg = f'Invalid colormap "{cmap}"'
                    raise ValueError(msg)
                cmap_obj = getattr(colormaps, cmap)
            else:
                try:
                    cmap_obj = colormaps[cmap]
                except KeyError:
                    msg = f"Invalid colormap '{cmap}'"
                    raise ValueError(msg) from None

    else:  # input is a list
        for item in cmap:
            if not isinstance(item, str):
                msg = 'When inputting a list as a cmap, each item should be a string.'  # type: ignore[unreachable]
                raise TypeError(msg)

        cmap_obj = ListedColormap(cmap)

    return cmap_obj


def get_default_cycler():
    """Return the default color cycler (matches matplotlib's default).

    Returns
    -------
    cycler.Cycler
        A cycler object for color that matches matplotlib's default colors.

    """
    return cycler('color', matplotlib_default_colors)


def get_hexcolors_cycler():
    """Return a color cycler for all of the available hexcolors.

    See ``pyvista.plotting.colors.hexcolors``.

    Returns
    -------
    cycler.Cycler
        A cycler object for color using all the available hexcolors from
        ``pyvista.plotting.colors.hexcolors``.

    """
    return cycler('color', hexcolors.keys())


def get_matplotlib_theme_cycler():
    """Return the color cycler of the current matplotlib theme.

    Returns
    -------
    cycler.Cycler
        Color cycler of the current matplotlib theme.

    """
    return plt.rcParams['axes.prop_cycle']


def color_scheme_to_cycler(scheme):
    """Convert a color scheme to a Cycler.

    Parameters
    ----------
    scheme : str | int | :vtk:`vtkColorSeries`
        Color scheme to be converted. If a string, it should correspond to a
        valid color scheme name (e.g., 'viridis'). If an integer, it should
        correspond to a valid color scheme ID. If an instance of
        :vtk:`vtkColorSeries`, it should be a valid color series.

    Returns
    -------
    cycler.Cycler
        A Cycler object with the color scheme.

    Raises
    ------
    ValueError
        If the provided `scheme` is not a valid color scheme.

    """
    if not isinstance(scheme, _vtk.vtkColorSeries):
        series = _vtk.vtkColorSeries()
        if isinstance(scheme, str):
            series.SetColorScheme(COLOR_SCHEMES.get(scheme.lower())['id'])  # type: ignore[index]
        elif isinstance(scheme, int):
            series.SetColorScheme(scheme)
        else:
            msg = f'Color scheme not understood: {scheme}'
            raise TypeError(msg)
    else:
        series = scheme
    colors = (series.GetColor(i) for i in range(series.GetNumberOfColors()))
    return cycler('color', colors)


def get_cycler(color_cycler):
    """Return a color cycler based on the input value.

    Parameters
    ----------
    color_cycler : str, list, tuple, or Cycler
        Specifies the desired color cycler. The value must be one of the following:
        - A list or tuple of color-like objects
        - A Cycler object with color-like objects
        - One of the following string values:
            - ``'default'``: Use the default color cycler (matches matplotlib's default)
            - ``'matplotlib'``: Dynamically get matplotlib's current theme's color cycler.
            - ``'all'``: Cycle through all available colors in
              ``pyvista.plotting.colors.hexcolors``
        - A named color scheme from ``pyvista.plotting.colors.COLOR_SCHEMES``

    Returns
    -------
    Cycler
        The color cycler corresponding to the input value.

    Raises
    ------
    ValueError
        Raised if the input is a string not found in named color schemes.
    TypeError
        Raised if the input is of an unsupported type.

    """
    if color_cycler is None:
        cycler_ = None
    elif isinstance(color_cycler, str):
        if color_cycler == 'default':
            cycler_ = get_default_cycler()
        elif color_cycler == 'matplotlib':
            cycler_ = get_matplotlib_theme_cycler()
        elif color_cycler == 'all':
            cycler_ = get_hexcolors_cycler()
        elif color_cycler in COLOR_SCHEMES:
            cycler_ = color_scheme_to_cycler(color_cycler)
        else:
            msg = f'color cycler of name `{color_cycler}` not found.'
            raise ValueError(msg)
    elif isinstance(color_cycler, (tuple, list)):
        cycler_ = cycler('color', color_cycler)
    elif isinstance(color_cycler, Cycler):
        cycler_ = color_cycler
    else:
        msg = f'color cycler of type {type(color_cycler)} not supported.'
        raise TypeError(msg)
    return cycler_



"""Module containing pyvista implementation of :vtk:`vtkLight`."""

from __future__ import annotations

from enum import IntEnum

import numpy as np

# imports here rather than in _vtk to avoid circular imports
try:
    from vtkmodules.vtkCommonMath import vtkMatrix4x4
    from vtkmodules.vtkRenderingCore import vtkLight
    from vtkmodules.vtkRenderingCore import vtkLightActor
except ImportError:  # pragma: no cover
    from vtk import vtkLight
    from vtk import vtkLightActor
    from vtk import vtkMatrix4x4

from typing import TYPE_CHECKING

from pyvista._deprecate_positional_args import _deprecate_positional_args
from pyvista.core import _validation
from pyvista.core._vtk_utilities import DisableVtkSnakeCase
from pyvista.core.utilities.arrays import vtkmatrix_from_array
from pyvista.core.utilities.misc import _NoNewAttrMixin

from .colors import Color

if TYPE_CHECKING:
    from pyvista.core._typing_core import TransformLike

    from ._typing import ColorLike


class LightType(IntEnum):
    """An enumeration for the light types."""

    HEADLIGHT = 1
    CAMERA_LIGHT = 2
    SCENE_LIGHT = 3

    def __str__(self):
        """Pretty name for a light type."""
        return self.name.replace('_', ' ').title()


class Light(_NoNewAttrMixin, DisableVtkSnakeCase, vtkLight):
    """Light class.

    Parameters
    ----------
    position : sequence[float], optional
        The position of the light. The interpretation of the position
        depends on the type of the light and whether the light has a
        transformation matrix.  See also the :py:attr:`position`
        property.

    focal_point : sequence[float], optional
        The focal point of the light. The interpretation of the focal
        point depends on the type of the light and whether the light
        has a transformation matrix.  See also the
        :py:attr:`focal_point` property.

    color : ColorLike, optional
        The color of the light. The ambient, diffuse and specular
        colors will all be set to this color on creation.

    light_type : str | int, default: 'scene light'
        The type of the light. If a string, one of ``'headlight'``,
        ``'camera light'`` or ``'scene light'``. If an int, one of 1,
        2 or 3, respectively. The class constants ``Light.HEADLIGHT``,
        ``Light.CAMERA_LIGHT`` and ``Light.SCENE_LIGHT`` are also
        available, respectively.

            - A headlight is attached to the camera, looking at its
              focal point along the axis of the camera.

            - A camera light also moves with the camera, but it can
              occupy a general position with respect to it.

            - A scene light is stationary with respect to the scene,
              as it does not follow the camera. This is the default.

    intensity : float, optional
        The brightness of the light (between 0 and 1).

    positional : bool, optional
        Set if the light is positional.

        The default is a directional light, i.e. an infinitely distant
        point source. A positional light with a cone angle of at least
        90 degrees acts like a spherical point source. A positional
        light with a cone angle that is less than 90 degrees is known
        as a spotlight.

    cone_angle : float, optional
        Cone angle of a positional light in degrees.

    show_actor : bool, default: False
        Show an actor for a spotlight that depicts the geometry of the
        beam.

    exponent : float, optional
        The exponent of the cosine used for spotlights. See also the
        :py:attr:`exponent` property.

    shadow_attenuation : float, optional
        The value of shadow attenuation.

        By default a light will be completely blocked when in shadow.
        By setting this value to less than 1.0 you can control how
        much light is attenuated when in shadow. Note that changing
        the :py:attr:`attenuation_values` of the light can make it pass
        through objects even if its shadow attenuation is 1.

    attenuation_values : sequence, optional
        Quadratic attenuation constants.

        The values are a 3-length sequence which specifies the constant,
        linear and quadratic constants in this order. These parameters
        only have an effect for positional lights.

    Examples
    --------
    Create a light at (10, 10, 10) and set its diffuse color to red.

    >>> import pyvista as pv
    >>> light = pv.Light(position=(10, 10, 10))
    >>> light.diffuse_color = 1.0, 0.0, 0.0

    Create a positional light at (0, 0, 3) with a cone angle of
    30, exponent of 20, and a visible actor.

    >>> light = pv.Light(
    ...     position=(0, 0, 3),
    ...     show_actor=True,
    ...     positional=True,
    ...     cone_angle=30,
    ...     exponent=20,
    ... )

    """

    # pull in light type enum values as class constants
    HEADLIGHT = LightType.HEADLIGHT
    CAMERA_LIGHT = LightType.CAMERA_LIGHT
    SCENE_LIGHT = LightType.SCENE_LIGHT

    @_deprecate_positional_args
    def __init__(  # noqa: PLR0917
        self,
        position=None,
        focal_point=None,
        color=None,
        light_type='scene light',
        intensity=None,
        positional=None,
        cone_angle=None,
        show_actor=False,  # noqa: FBT002
        exponent=None,
        shadow_attenuation=None,
        attenuation_values=None,
    ):
        """Initialize the light."""
        super().__init__()
        self._renderers = []
        self.actor = None

        if position is not None:
            self.position = position
        if focal_point is not None:
            self.focal_point = focal_point

        if color is not None:
            self.ambient_color = color
            self.diffuse_color = color
            self.specular_color = color

        if isinstance(light_type, str):
            # be forgiving: ignore spaces and case
            light_type_orig = light_type
            type_normalized = light_type.replace(' ', '').lower()
            mapping = {
                'headlight': LightType.HEADLIGHT,
                'cameralight': LightType.CAMERA_LIGHT,
                'scenelight': LightType.SCENE_LIGHT,
            }
            try:
                light_type = mapping[type_normalized]
            except KeyError:
                light_keys = ', '.join(mapping)
                msg = (
                    f'Invalid light_type "{light_type_orig}"\n'
                    f'Choose from one of the following: {light_keys}'
                )
                raise ValueError(msg) from None
        elif not isinstance(light_type, int):
            msg = f'Parameter light_type must be int or str, not {type(light_type).__name__}.'
            raise TypeError(msg)
        # LightType is an int subclass

        self.light_type = light_type

        if intensity is not None:
            self.intensity = intensity

        if cone_angle is not None:
            self.cone_angle = cone_angle

        if exponent is not None:
            self.exponent = exponent

        if positional is not None:
            self.positional = positional

        if shadow_attenuation is not None:
            self.shadow_attenuation = shadow_attenuation

        if attenuation_values is not None:
            self.attenuation_values = attenuation_values

        # Add the light actor
        self.actor = vtkLightActor()
        self.actor.SetLight(self)
        self.actor.SetVisibility(show_actor)

    def __repr__(self) -> str:
        """Print a repr specifying the id of the light and its light type."""
        return f'<{self.__class__.__name__} ({self.light_type}) at {hex(id(self))}>'

    def __eq__(self, other) -> bool:
        """Compare whether the relevant attributes of two lights are equal."""
        # attributes which are native python types and thus implement __eq__
        native_attrs = [
            'light_type',
            'position',
            'focal_point',
            'ambient_color',
            'diffuse_color',
            'specular_color',
            'intensity',
            'on',
            'positional',
            'exponent',
            'cone_angle',
            'attenuation_values',
            'shadow_attenuation',
        ]
        for attr in native_attrs:
            if getattr(self, attr) != getattr(other, attr):
                return False

        # check transformation matrix element by element (if it exists)
        this_trans = self.transform_matrix
        that_trans = other.transform_matrix
        trans_count = sum(1 for trans in [this_trans, that_trans] if trans is not None)
        if trans_count == 1:
            # either but not both are None
            return False
        if trans_count == 2:
            for i in range(4):
                for j in range(4):
                    if this_trans.GetElement(i, j) != that_trans.GetElement(i, j):
                        return False
        return True

    __hash__ = None  # type: ignore[assignment]  # https://github.com/pyvista/pyvista/pull/7671

    def __del__(self):
        """Clean up when the light is being destroyed."""
        self.actor = None
        self._renderers.clear()

    @property
    def shadow_attenuation(self):  # numpydoc ignore=RT01
        """Return or set the value of shadow attenuation.

        By default a light will be completely blocked when in shadow.
        By setting this value to less than 1.0 you can control how
        much light is attenuated when in shadow. Note that changing
        the :py:attr:`attenuation_values` of the light can make it pass
        through objects even if its shadow attenuation is 1.

        Examples
        --------
        Set the shadow attenuation to 0.5

        >>> import pyvista as pv
        >>> light = pv.Light()
        >>> light.shadow_attenuation = 0.5
        >>> light.shadow_attenuation
        0.5

        """
        return self.GetShadowAttenuation()

    @shadow_attenuation.setter
    def shadow_attenuation(self, value):
        self.SetShadowAttenuation(value)

    @property
    def ambient_color(self):  # numpydoc ignore=RT01
        """Return or set the ambient color of the light.

        When setting, the color must be a 3-length sequence or a string.
        For example:

            * ``color='white'``
            * ``color='w'``
            * ``color=[1.0, 1.0, 1.0]``
            * ``color='#FFFFFF'``

        Examples
        --------
        Create a light and set its ambient color to red.

        >>> import pyvista as pv
        >>> light = pv.Light()
        >>> light.ambient_color = 'red'
        >>> light.ambient_color
        Color(name='red', hex='#ff0000ff', opacity=255)

        """
        return Color(self.GetAmbientColor())

    @ambient_color.setter
    def ambient_color(self, color: ColorLike):
        self.SetAmbientColor(Color(color).float_rgb)

    @property
    def diffuse_color(self):  # numpydoc ignore=RT01
        """Return or set the diffuse color of the light.

        When setting, the color must be a 3-length sequence or a string.
        For example:

            * ``color='white'``
            * ``color='w'``
            * ``color=[1.0, 1.0, 1.0]``
            * ``color='#FFFFFF'``

        Examples
        --------
        Create a light and set its diffuse color to blue.

        >>> import pyvista as pv
        >>> light = pv.Light()
        >>> light.diffuse_color = (0.0, 0.0, 1.0)
        >>> light.diffuse_color
        Color(name='blue', hex='#0000ffff', opacity=255)

        """
        return Color(self.GetDiffuseColor())

    @diffuse_color.setter
    def diffuse_color(self, color: ColorLike):
        self.SetDiffuseColor(Color(color).float_rgb)

    @property
    def specular_color(self):  # numpydoc ignore=RT01
        """Return or set the specular color of the light.

        When setting, the color must be a 3-length sequence or a string.
        For example:

            * ``color='white'``
            * ``color='w'``
            * ``color=[1.0, 1.0, 1.0]``
            * ``color='#FFFFFF'``

        Examples
        --------
        Create a light and set its specular color to bright green.

        >>> import pyvista as pv
        >>> light = pv.Light()
        >>> light.specular_color = '#00FF00'
        >>> light.specular_color
        Color(name='lime', hex='#00ff00ff', opacity=255)

        """
        return Color(self.GetSpecularColor())

    @specular_color.setter
    def specular_color(self, color: ColorLike):
        self.SetSpecularColor(Color(color).float_rgb)

    @property
    def position(self):  # numpydoc ignore=RT01
        """Return the position of the light.

        Note: the position is defined in the coordinate space
        indicated by the light's transformation matrix (if it
        exists). To get the light's world space position, use the
        (read-only) :py:attr:`world_position` property.

        Examples
        --------
        Create a light positioned at ``(10, 10, 10)`` after
        initialization, and note how the position is unaffected by a
        non-trivial transform matrix.

        >>> import numpy as np
        >>> import pyvista as pv
        >>> light = pv.Light()
        >>> light.position = (10, 10, 10)
        >>> # set a "random" transformation matrix
        >>> light.transform_matrix = np.arange(4 * 4).reshape(4, 4)
        >>> light.position
        (10.0, 10.0, 10.0)

        """
        return self.GetPosition()

    @position.setter
    def position(self, pos):
        self.SetPosition(pos)

    @property
    def world_position(self):  # numpydoc ignore=RT01
        """Return the world space position of the light.

        The world space position is the :py:attr:`position` property
        transformed by the light's transform matrix if it exists. The
        value of this read-only property corresponds to the
        :vtk:`vtkLight.GetTransformedPosition` method.

        Examples
        --------
        Create a light with a transformation matrix that corresponds to a
        90-degree rotation around the z-axis and a shift by (0, 0, -1), and
        check that the light's position transforms as expected.

        >>> import numpy as np
        >>> import pyvista as pv
        >>> light = pv.Light(position=(1, 0, 3))
        >>> trans = np.zeros((4, 4))
        >>> trans[:-1, :-1] = [[0, -1, 0], [1, 0, 0], [0, 0, 1]]
        >>> trans[:-1, -1] = [0, 0, -1]
        >>> light.transform_matrix = trans
        >>> light.position
        (1.0, 0.0, 3.0)
        >>> light.world_position
        (0.0, 1.0, 2.0)

        """
        return self.GetTransformedPosition()

    @property
    def focal_point(self):  # numpydoc ignore=RT01
        """Return the focal point of the light.

        Note: the focal point is defined in the coordinate space
        indicated by the light's transformation matrix (if it
        exists). To get the light's world space focal point, use the
        (read-only) :py:attr:`world_focal_point` property.

        Examples
        --------
        Create a light at (10, 10, 10) shining at (0, 0, 1).

        >>> import pyvista as pv
        >>> light = pv.Light(position=(10, 10, 10))
        >>> light.focal_point = (0, 0, 1)

        """
        return self.GetFocalPoint()

    @focal_point.setter
    def focal_point(self, pos):
        self.SetFocalPoint(pos)

    @property
    def world_focal_point(self):  # numpydoc ignore=RT01
        """Return the world space focal point of the light.

        The world space focal point is the :py:attr:`focal_point`
        property transformed by the light's transform matrix if it
        exists. The value of this read-only property corresponds to
        the :vtk:`vtkLight.GetTransformedFocalPoint` method.

        Examples
        --------
        Create a light with a transformation matrix that corresponds
        to a 90-degree rotation around the z-axis and a shift by (0,
        0, -1), and check that the light's focal point transforms as
        expected.

        >>> import numpy as np
        >>> import pyvista as pv
        >>> light = pv.Light(focal_point=(1, 0, 3))
        >>> trans = np.zeros((4, 4))
        >>> trans[:-1, :-1] = [[0, -1, 0], [1, 0, 0], [0, 0, 1]]
        >>> trans[:-1, -1] = [0, 0, -1]
        >>> light.transform_matrix = trans
        >>> light.focal_point
        (1.0, 0.0, 3.0)
        >>> light.world_focal_point
        (0.0, 1.0, 2.0)

        """
        return self.GetTransformedFocalPoint()

    @property
    def intensity(self):  # numpydoc ignore=RT01
        """Return or set the brightness of the light (between 0 and 1).

        Examples
        --------
        Light the two sides of a cube with lights of different
        brightness.

        >>> import pyvista as pv
        >>> pl = pv.Plotter(lighting='none')
        >>> _ = pl.add_mesh(pv.Cube(), color='cyan')
        >>> light_bright = pv.Light(position=(3, 0, 0), light_type='scene light')
        >>> light_dim = pv.Light(position=(0, 3, 0), light_type='scene light')
        >>> light_dim.intensity = 0.5
        >>> for light in light_bright, light_dim:
        ...     light.positional = True
        ...     pl.add_light(light)
        >>> pl.show()

        """
        return self.GetIntensity()

    @intensity.setter
    def intensity(self, intensity):
        self.SetIntensity(intensity)

    @property
    def on(self):  # numpydoc ignore=RT01
        """Return or set whether the light is on.

        This corresponds to the Switch state of the :vtk:`vtkLight` class.

        Examples
        --------
        Create a light, check if it's on by default, and turn it off.

        >>> import pyvista as pv
        >>> light = pv.Light()
        >>> light.on
        True
        >>> light.on = False

        """
        return bool(self.GetSwitch())

    @on.setter
    def on(self, state):
        self.SetSwitch(state)

    @property
    def positional(self):  # numpydoc ignore=RT01
        """Return or set whether the light is positional.

        The default is a directional light, i.e. an infinitely distant
        point source. A positional light with a cone angle of at least
        90 degrees acts like a spherical point source. A positional
        light with a cone angle that is less than 90 degrees is known
        as a spotlight.

        Attenuation and cone angles are only used for positional
        lights.  The :py:attr:`exponent` property is only used for
        spotlights.  Positional lights with a cone angle of at least
        90 degrees don't show angular dependence of their beams, but
        they display attenuation.

        If the light is changed to directional, its actor (if previously
        shown) is automatically hidden.

        Examples
        --------
        Create a spotlight shining on the origin.

        >>> import pyvista as pv
        >>> light = pv.Light(position=(1, 1, 1))
        >>> light.positional = True
        >>> light.cone_angle = 30

        """
        return bool(self.GetPositional())

    @positional.setter
    def positional(self, state):
        if not state:
            self.hide_actor()
        self.SetPositional(state)
        self._check_actor()

    def _check_actor(self):
        """Check if the light actor should be added or removed from attached renderers.

        This should be called whenever positional state or cone angle
        are changed.

        """
        if self.actor is None:
            # only should occur on __init__
            return

        actor_state = self.cone_angle < 90 and self.positional
        actor_name = self.actor.GetAddressAsString('')

        # add or remove the actor from the renderer
        for renderer in self._renderers:
            if actor_state:
                if actor_name not in renderer.actors:
                    renderer.add_actor(self.actor, render=False)
            elif actor_name in renderer.actors:
                renderer.remove_actor(self.actor, render=False)

    @property
    def exponent(self):  # numpydoc ignore=RT01
        """Return or set the exponent of the cosine used for spotlights.

        With a spotlight (a positional light with cone angle less than
        90 degrees) the shape of the light beam within the light cone
        varies with the angle from the light's axis, and the variation
        of the intensity depends as the cosine of this angle raised to
        an exponent, which is 1 by default. Increasing the exponent
        makes the beam sharper (more focused around the axis),
        decreasing it spreads the beam out.

        Note that since the angular dependence defined by this
        property and the truncation performed by the
        :py:attr:`cone_angle` are independent, for spotlights with
        narrow beams (small :py:attr:`cone_angle`) it is harder to see
        the angular variation of the intensity, and a lot higher
        exponent might be necessary to visibly impact the angular
        distribution of the beam.

        Examples
        --------
        Plot three planes lit by three spotlights with exponents of 1,
        2 and 5.  The one with the lowest exponent has the broadest
        beam.

        >>> import pyvista as pv
        >>> pl = pv.Plotter(lighting='none')
        >>> for offset, exponent in zip([0, 1.5, 3], [1, 2, 5]):
        ...     _ = pl.add_mesh(pv.Plane(center=(offset, 0, 0)), color='white')
        ...     light = pv.Light(
        ...         position=(offset, 0, 0.1),
        ...         focal_point=(offset, 0, 0),
        ...     )
        ...     light.exponent = exponent
        ...     light.positional = True
        ...     light.cone_angle = 80
        ...     pl.add_light(light)
        >>> pl.view_xy()
        >>> pl.show()

        """
        return self.GetExponent()

    @exponent.setter
    def exponent(self, exp):
        self.SetExponent(exp)

    @property
    def cone_angle(self):  # numpydoc ignore=RT01
        """Return or set the cone angle of a positional light.

        The angle is in degrees and is measured between the axis of
        the cone and an extremal ray of the cone. A value smaller than
        90 has spot lighting effects, anything equal to and above 90
        is just a positional light, i.e. a spherical point source.

        Regarding the angular distribution of the light, the cone
        angle merely truncates the beam, the shape of which is defined
        by the :py:attr:`exponent`.  If the cone angle is at least 90
        degrees then there is no angular dependence.

        If the light's cone angle is increased to 90 degrees or above,
        its actor (if previously shown) is automatically hidden.

        See Also
        --------
        :ref:`beam_shape_example`

        Examples
        --------
        Plot three planes lit by three spotlights with varying cone
        angles.  Use a large exponent to cause a visible angular
        variation of the intensity of the beams.

        >>> import pyvista as pv
        >>> pl = pv.Plotter(lighting='none')
        >>> for offset, angle in zip([0, 1.5, 3], [70, 30, 20]):
        ...     _ = pl.add_mesh(pv.Plane(center=(offset, 0, 0)), color='white')
        ...     light = pv.Light(position=(offset, 0, 1), focal_point=(offset, 0, 0))
        ...     light.exponent = 15
        ...     light.positional = True
        ...     light.cone_angle = angle
        ...     pl.add_light(light)
        >>> pl.view_xy()
        >>> pl.show()

        """
        return self.GetConeAngle()

    @cone_angle.setter
    def cone_angle(self, angle):
        if angle >= 90:
            self.hide_actor()
        self.SetConeAngle(angle)
        self._check_actor()

    @property
    def attenuation_values(self):  # numpydoc ignore=RT01
        """Return or set the quadratic attenuation constants.

        The values are 3-length sequences which specify the constant,
        linear and quadratic constants in this order. These parameters
        only have an effect for positional lights.

        Attenuation refers to the dampening of a beam of light as it
        gets further away from the point source. The three constants
        describe three different profiles for dampening with
        distance. A larger attenuation constant corresponds to more
        rapid decay with distance.

        See Also
        --------
        :ref:`attenuation_example`

        Examples
        --------
        Plot three cubes lit by two lights with different attenuation
        profiles.  The blue light has slower linear attenuation, the
        green one has quadratic attenuation that makes it decay
        faster. Note that there are no shadow effects included so each
        box gets lit by both lights.

        >>> import pyvista as pv
        >>> pl = pv.Plotter(lighting='none')
        >>> for offset in 1, 2.5, 4:
        ...     _ = pl.add_mesh(pv.Cube(center=(offset, offset, 0)), color='white')
        >>> colors = ['b', 'g']
        >>> all_attenuations = [(0, 0.1, 0), (0, 0, 0.1)]
        >>> centers = [(0, 1, 0), (1, 0, 0)]
        >>> for color, attenuation_constants, center in zip(
        ...     colors, all_attenuations, centers
        ... ):
        ...     light = pv.Light(position=center, color=color)
        ...     light.focal_point = (1 + center[0], 1 + center[1], 0)
        ...     light.cone_angle = 90
        ...     light.positional = True
        ...     light.attenuation_values = attenuation_constants
        ...     pl.add_light(light)
        >>> pl.view_vector((-1, -1, 1))
        >>> pl.show()

        """
        return self.GetAttenuationValues()

    @attenuation_values.setter
    def attenuation_values(self, values):
        self.SetAttenuationValues(values)

    @property
    def transform_matrix(self):  # numpydoc ignore=RT01
        """Return (if any) or set the transformation matrix of the light.

        The transformation matrix is ``None`` by default, and it is
        stored as a :vtk:`vtkMatrix4x4` object when set. If set, the
        light's parameters (position and focal point) are transformed
        by the matrix before being rendered. See also the
        :py:attr:`world_position` and :py:attr:`world_focal_point`
        read-only properties that can differ from :py:attr:`position`
        and :py:attr:`focal_point`, respectively.

        The 4-by-4 transformation matrix is a tool to encode a general
        linear transformation and a translation (an affine
        transform). The 3-by-3 principal submatrix (the top left
        corner of the matrix) encodes a three-dimensional linear
        transformation (e.g. some rotation around the origin). The top
        three elements in the last column of the matrix encode a
        three-dimensional translation. The last row of the matrix is
        redundant.

        Examples
        --------
        Create a light with a transformation matrix that corresponds
        to a 90-degree rotation around the z-axis and a shift by (0,
        0, -1), and check that the light's position transforms as
        expected.

        >>> import numpy as np
        >>> import pyvista as pv
        >>> light = pv.Light(position=(1, 0, 3))
        >>> trans = np.zeros((4, 4))
        >>> trans[:-1, :-1] = [[0, -1, 0], [1, 0, 0], [0, 0, 1]]
        >>> trans[:-1, -1] = [0, 0, -1]
        >>> light.transform_matrix = trans
        >>> light.position
        (1.0, 0.0, 3.0)
        >>> light.world_position
        (0.0, 1.0, 2.0)

        """
        return self.GetTransformMatrix()

    @transform_matrix.setter
    def transform_matrix(self, matrix: TransformLike):
        if matrix is None or isinstance(matrix, vtkMatrix4x4):
            self.SetTransformMatrix(matrix)
        else:
            trans = _validation.validate_transform4x4(matrix)
            self.SetTransformMatrix(vtkmatrix_from_array(trans))

    @property
    def light_type(self):  # numpydoc ignore=RT01
        """Return or set the light type.

        The default light type is a scene light which lives in world
        coordinate space.

        A headlight is attached to the camera and always points at the
        camera's focal point.

        A camera light also moves with the camera, but it can have an
        arbitrary relative position to the camera. Camera lights are
        defined in a coordinate space where the camera is located at
        (0, 0, 1), looking towards (0, 0, 0) at a distance of 1, with
        up being (0, 1, 0). Camera lights use the transform matrix to
        establish this space, i.e. they have a fixed :py:attr:`position`
        with respect to the camera, and moving the camera only
        affects the :py:attr:`world_position` via changes in the
        :py:attr:`transform_matrix` (and the same goes for the focal
        point).

        The property returns class constant values from an enum:

            - ``Light.HEADLIGHT == 1``
            - ``Light.CAMERA_LIGHT == 2``
            - ``Light.SCENE_LIGHT == 3``

        If setting the value, either an integer code or a class constant enum
        value must be used.

        Examples
        --------
        Check the type of lights for the first two lights of the default
        light kit of plotters.

        >>> import pyvista as pv
        >>> pl = pv.Plotter()
        >>> lights = pl.renderer.lights[:2]
        >>> [light.light_type for light in lights]
        [<LightType.HEADLIGHT: 1>, <LightType.CAMERA_LIGHT: 2>]

        Change the light type of the default light kit's headlight to a scene light.

        >>> import pyvista as pv
        >>> pl = pv.Plotter()
        >>> lights = pl.renderer.lights[:2]
        >>> lights[0].light_type = pv.Light.SCENE_LIGHT
        >>> [light.light_type for light in lights]
        [<LightType.SCENE_LIGHT: 3>, <LightType.CAMERA_LIGHT: 2>]

        """
        return LightType(self.GetLightType())

    @light_type.setter
    def light_type(self, ltype):
        if not isinstance(ltype, int):
            # note that LightType is an int subclass
            msg = f'Light type must be an integer subclass instance, got {ltype} instead.'
            raise TypeError(msg)
        self.SetLightType(ltype)

    @property
    def is_headlight(self):  # numpydoc ignore=RT01
        """Return whether the light is a headlight.

        Examples
        --------
        Verify that the first light of the default light kit is a headlight.

        >>> import pyvista as pv
        >>> pl = pv.Plotter()
        >>> lights = pl.renderer.lights
        >>> [light.is_headlight for light in lights]
        [True, False, False, False, False]

        """
        return bool(self.LightTypeIsHeadlight())

    @property
    def is_camera_light(self):  # numpydoc ignore=RT01
        """Return whether the light is a camera light.

        Examples
        --------
        Verify that four out of five lights of the default light kit
        are camera lights.

        >>> import pyvista as pv
        >>> pl = pv.Plotter()
        >>> lights = pl.renderer.lights
        >>> [light.is_camera_light for light in lights]
        [False, True, True, True, True]

        """
        return bool(self.LightTypeIsCameraLight())

    @property
    def is_scene_light(self):  # numpydoc ignore=RT01
        """Return whether the light is a scene light.

        Examples
        --------
        Verify that none of the lights of the default light kit are
        scene lights.

        >>> import pyvista as pv
        >>> pl = pv.Plotter()
        >>> lights = pl.renderer.lights
        >>> [light.is_scene_light for light in lights]
        [False, False, False, False, False]

        """
        return bool(self.LightTypeIsSceneLight())

    def switch_on(self):
        """Switch on the light.

        Examples
        --------
        Create a light, switch it off and switch it back on again.

        >>> import pyvista as pv
        >>> light = pv.Light()
        >>> light.on = False
        >>> light.switch_on()

        """
        self.SwitchOn()

    def switch_off(self):
        """Switch off the light.

        Examples
        --------
        Create a light and switch it off.

        >>> import pyvista as pv
        >>> light = pv.Light()
        >>> light.switch_off()

        """
        self.SwitchOff()

    def set_direction_angle(self, elev, azim):
        """Set the position and focal point of a directional light.

        The light is switched to directional (non-positional). The
        focal point is set to the origin. The position is defined in
        terms of an elevation and an azimuthal angle, both in degrees.

        Note that the equivalent :vtk:`vtkLight.SetDirectionAngle` method
        uses a surprising coordinate system where the (x', y', z') axes of
        the method correspond to the (z, x, y) axes of the renderer.
        This method reimplements the functionality in a way that ``elev``
        is the conventional elevation and ``azim`` is the conventional azimuth.
        In particular:

          * ``elev = 0``, ``azim = 0`` is the +x direction
          * ``elev = 0``, ``azim = 90`` is the +y direction
          * ``elev = 90``, ``azim = 0`` is the +z direction

        Parameters
        ----------
        elev : float
            The elevation of the directional light.

        azim : float
            The azimuthal angle of the directional light.

        Examples
        --------
        Create a light that shines on the origin from a 30-degree
        elevation in the xz plane.

        >>> import pyvista as pv
        >>> light = pv.Light()
        >>> light.set_direction_angle(30, 0)

        """
        self.positional = False
        self.focal_point = (0, 0, 0)
        theta = np.radians(90 - elev)
        phi = np.radians(azim)
        self.position = (np.sin(theta) * np.cos(phi), np.sin(theta) * np.sin(phi), np.cos(theta))

    @_deprecate_positional_args
    def copy(self, deep=True):  # noqa: FBT002
        """Return a shallow or a deep copy of the light.

        The only mutable attribute of :class:`pyvista.Light` is the
        transformation matrix (if it exists). Thus asking for a
        shallow copy merely implies that the returned light and the
        original share the transformation matrix instance.

        Parameters
        ----------
        deep : bool, default: True
            Whether to return a deep copy rather than a shallow
            one.

        Returns
        -------
        pyvista.Light
            Copied light.

        Examples
        --------
        Create a light and check that it shares a transformation
        matrix with its shallow copy.

        >>> import pyvista as pv
        >>> light = pv.Light()
        >>> light.transform_matrix = [
        ...     [1, 0, 0, 0],
        ...     [0, 1, 0, 0],
        ...     [0, 0, 1, 0],
        ...     [0, 0, 0, 1],
        ... ]
        >>> shallow_copied = light.copy(deep=False)
        >>> shallow_copied == light
        True
        >>> shallow_copied.transform_matrix is light.transform_matrix
        True

        """
        immutable_attrs = [
            'light_type',
            'position',
            'focal_point',
            'ambient_color',
            'diffuse_color',
            'specular_color',
            'intensity',
            'on',
            'positional',
            'exponent',
            'cone_angle',
            'attenuation_values',
            'shadow_attenuation',
        ]
        new_light = Light()

        for attr in immutable_attrs:
            value = getattr(self, attr)
            setattr(new_light, attr, value)

        if deep and self.transform_matrix is not None:
            new_light.transform_matrix = vtkMatrix4x4()
            new_light.transform_matrix.DeepCopy(self.transform_matrix)
        else:
            new_light.transform_matrix = self.transform_matrix

        # light actors are private, always copy, but copy visibility state as well
        new_light.actor.SetVisibility(self.actor.GetVisibility())  # type: ignore[union-attr]

        return new_light

    def set_headlight(self):
        """Set the light to be a headlight.

        Headlights are fixed to the camera and always point to the
        focal point of the camera. Calling this method will reset the
        light's transformation matrix.

        Examples
        --------
        >>> import pyvista as pv
        >>> light = pv.Light()
        >>> light.set_headlight()
        >>> light.light_type
        <LightType.HEADLIGHT: 1>

        """
        self.SetLightTypeToHeadlight()

    def set_camera_light(self):
        """Set the light to be a camera light.

        A camera light moves with the camera, but it can have an
        arbitrary relative position to the camera. Camera lights are
        defined in a coordinate space where the camera is located at
        (0, 0, 1), looking towards (0, 0, 0) at a distance of 1, with
        up being (0, 1, 0).  Camera lights use the transformation
        matrix to establish this space.  Calling this method will
        reset the light's transformation matrix.

        Examples
        --------
        >>> import pyvista as pv
        >>> light = pv.Light()
        >>> light.set_camera_light()
        >>> light.light_type
        <LightType.CAMERA_LIGHT: 2>

        """
        self.SetLightTypeToCameraLight()

    def set_scene_light(self):
        """Set the light to be a scene light.

        Scene lights are stationary with respect to the scene.
        Calling this method will reset the light's transformation
        matrix.

        Examples
        --------
        >>> import pyvista as pv
        >>> light = pv.Light()
        >>> light.set_scene_light()
        >>> light.light_type
        <LightType.SCENE_LIGHT: 3>

        """
        self.SetLightTypeToSceneLight()

    @classmethod
    def from_vtk(cls, vtk_light):
        """Create a light from a :vtk:`vtkLight`, resulting in a copy.

        Parameters
        ----------
        vtk_light : :vtk:`vtkLight`
            The :vtk:`vtkLight` to be copied.

        Returns
        -------
        pyvista.Light
            Wrapped light.

        """
        if not isinstance(vtk_light, vtkLight):
            msg = f'Expected vtk.vtkLight object, got {type(vtk_light).__name__} instead.'
            raise TypeError(msg)

        light = cls()
        light.light_type = vtk_light.GetLightType()  # resets transformation matrix
        light.position = vtk_light.GetPosition()
        light.focal_point = vtk_light.GetFocalPoint()
        light.ambient_color = vtk_light.GetAmbientColor()
        light.diffuse_color = vtk_light.GetDiffuseColor()
        light.specular_color = vtk_light.GetSpecularColor()
        light.intensity = vtk_light.GetIntensity()
        light.on = vtk_light.GetSwitch()
        light.positional = vtk_light.GetPositional()
        light.exponent = vtk_light.GetExponent()
        light.cone_angle = vtk_light.GetConeAngle()
        light.attenuation_values = vtk_light.GetAttenuationValues()
        trans = vtk_light.GetTransformMatrix()
        light.transform_matrix = trans
        light.shadow_attenuation = vtk_light.GetShadowAttenuation()

        return light

    def show_actor(self):
        """Show an actor for a spotlight that depicts the geometry of the beam.

        For a directional light or a positional light with
        :py:attr:`cone_angle` of at least 90 degrees the method
        doesn't do anything. If the light is changed so that it
        becomes a spotlight, this method has to be called again for
        the actor to show. To hide the actor see :func:`hide_actor`.

        Examples
        --------
        Create a scene containing a cube lit with a cyan spotlight and
        visualize the light using an actor.

        >>> import pyvista as pv
        >>> pl = pv.Plotter()
        >>> _ = pl.add_mesh(pv.Cube(), color='white')
        >>> for light in pl.renderer.lights:
        ...     light.intensity /= 5
        >>> spotlight = pv.Light(position=(-1, 1, 1), color='cyan')
        >>> spotlight.positional = True
        >>> spotlight.cone_angle = 20
        >>> spotlight.intensity = 10
        >>> spotlight.exponent = 40
        >>> spotlight.show_actor()
        >>> pl.add_light(spotlight)
        >>> pl.show()

        """
        if not self.positional or self.cone_angle >= 90:
            return
        self.actor.VisibilityOn()  # type: ignore[union-attr]

    def hide_actor(self):
        """Hide the actor for a positional light that depicts the geometry of the beam.

        For a directional light the function doesn't do anything.

        Examples
        --------
        >>> import pyvista as pv
        >>> light = pv.Light()
        >>> light.hide_actor()

        """
        if not self.positional:
            return
        self.actor.VisibilityOff()  # type: ignore[union-attr]

    @property
    def renderers(self):  # numpydoc ignore=RT01
        """Return the renderers associated with this light."""
        return self._renderers

    def add_renderer(self, renderer):
        """Attach a renderer to this light.

        Parameters
        ----------
        renderer : :vtk:`vtkRenderer`
            Renderer.

        """
        # quick check to avoid adding twice
        if renderer not in self.renderers:
            self.renderers.append(renderer)

        # verify that the renderer has the light actor if applicable
        self._check_actor()



from __future__ import annotations

import pathlib

import meshio
import numpy as np
import pytest

import pyvista as pv
from pyvista import examples


@pytest.fixture
def empty():
    return pv.UnstructuredGrid()


@pytest.fixture
def cow_ugrid():
    return examples.download_cow().cast_to_unstructured_grid()


@pytest.fixture
def points_only(cow_ugrid):
    cow_ugrid.cells = np.array((), dtype=int)
    assert cow_ugrid.n_cells == 0
    return cow_ugrid


@pytest.fixture
def airplane_ugrid(airplane):
    return airplane.cast_to_unstructured_grid()


@pytest.fixture
def uniform_ugrid(uniform):
    return uniform.cast_to_unstructured_grid()


@pytest.fixture
def uniform2d():
    return pv.ImageData(dimensions=(10, 10, 1)).cast_to_unstructured_grid()


@pytest.fixture
def hybrid():
    cells = [
        8, 0, 1, 2, 3, 4, 5, 6, 7,
        5, 4, 5, 6, 7, 8,
        4, 4, 8, 7, 9,
        4, 5, 6, 8, 10,
        6, 1, 11, 5, 2, 12, 6,
        6, 13, 0, 4, 14, 3, 7,
        30, 7,
        3, 5, 6, 10,
        3, 11, 12, 15,
        3, 5, 11, 15,
        3, 5, 15, 10,
        3, 6, 10, 15,
        3, 6, 15, 12,
        4, 11, 12, 6, 5,
        30, 7,
        3, 4, 7, 9,
        3, 13, 14, 16,
        3, 4, 16, 13,
        3, 4, 9, 16,
        3, 7, 9, 16,
        3, 7, 16, 14,
        4, 13, 14, 7, 4,
    ]  # fmt: skip
    return pv.UnstructuredGrid(
        cells,
        [
            pv.CellType.HEXAHEDRON,
            pv.CellType.PYRAMID,
            pv.CellType.TETRA,
            pv.CellType.TETRA,
            pv.CellType.WEDGE,
            pv.CellType.WEDGE,
            pv.CellType.POLYHEDRON,
            pv.CellType.POLYHEDRON,
        ],
        [
            [0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
            [1.0, 1.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 0.0, 1.0],
            [1.0, 0.0, 1.0],
            [1.0, 1.0, 1.0],
            [0.0, 1.0, 1.0],
            [0.5, 0.5, 1.5],
            [0.0, 0.5, 1.5],
            [1.0, 0.5, 1.5],
            [2.0, 0.0, 0.0],
            [2.0, 1.0, 0.0],
            [-1.0, 0.0, 0.0],
            [-1.0, 1.0, 0.0],
            [2.0, 0.5, 1.0],
            [-1.0, 0.5, 1.0],
        ],
    )


@pytest.fixture
def quad_pixel_hex_voxel():
    return (
        examples.cells.Quadrilateral()
        + examples.cells.Pixel()
        + examples.cells.Hexahedron()
        + examples.cells.Voxel()
    )


@pytest.fixture
def mesh2d():
    return meshio.Mesh(
        points=[[0.0, 0.0], [1.0, 0.0], [0.0, 1.0], [1.0, 1.0]],
        cells=[('triangle', [[0, 1, 2], [1, 3, 2]])],
        cell_sets={'tri1': [[0]], 'tri2': [[1]]},
    )


@pytest.fixture
def polyhedron():
    return meshio.Mesh(
        points=[
            [0.3568221, -0.49112344, 0.79465446],
            [-0.3568221, -0.49112344, 0.79465446],
            [0.3568221, 0.49112344, -0.79465446],
            [-0.3568221, 0.49112344, -0.79465446],
            [0.0, 0.98224693, 0.18759243],
            [0.0, 0.60706198, 0.79465446],
            [0.0, -0.60706198, -0.79465446],
            [0.0, -0.98224693, -0.18759243],
            [0.93417233, 0.30353101, 0.18759247],
            [0.93417233, -0.30353101, -0.18759247],
            [-0.93417233, 0.30353101, 0.18759247],
            [-0.93417233, -0.30353101, -0.18759247],
            [-0.57735026, 0.18759249, 0.79465446],
            [0.57735026, -0.79465446, 0.18759249],
            [-0.57735026, -0.18759249, -0.79465446],
            [0.57735026, 0.79465446, -0.18759249],
            [0.57735026, 0.18759249, 0.79465446],
            [-0.57735026, 0.79465446, -0.18759249],
            [-0.57735026, -0.79465446, 0.18759249],
            [0.57735026, -0.18759249, -0.79465446],
            [0.3568221, 0.49112344, -1.0],
            [0.57735026, -0.18759249, -1.0],
            [0.0, -0.60706198, -1.0],
            [-0.57735026, -0.18759249, -1.0],
            [-0.3568221, 0.49112344, -1.0],
            [0.3568221, -0.49112344, 1.0],
            [0.57735026, 0.18759249, 1.0],
            [0.0, 0.60706198, 1.0],
            [-0.57735026, 0.18759249, 1.0],
            [-0.3568221, -0.49112344, 1.0],
        ],
        cells=[
            (
                'polyhedron20',
                [
                    [
                        [0, 16, 5, 12, 1],
                        [1, 18, 7, 13, 0],
                        [2, 19, 6, 14, 3],
                        [3, 17, 4, 15, 2],
                        [4, 5, 16, 8, 15],
                        [5, 4, 17, 10, 12],
                        [6, 7, 18, 11, 14],
                        [7, 6, 19, 9, 13],
                        [8, 16, 0, 13, 9],
                        [9, 19, 2, 15, 8],
                        [10, 17, 3, 14, 11],
                        [11, 18, 1, 12, 10],
                    ],
                ],
            ),
            (
                'polyhedron10',
                [
                    [
                        [2, 19, 6, 14, 3],
                        [20, 21, 19, 2],
                        [21, 22, 6, 19],
                        [22, 23, 14, 6],
                        [23, 24, 3, 14],
                        [24, 20, 2, 3],
                        [20, 21, 22, 23, 24],
                    ],
                    [
                        [0, 16, 5, 12, 1],
                        [0, 16, 26, 25],
                        [16, 5, 27, 26],
                        [5, 12, 28, 27],
                        [12, 1, 29, 28],
                        [1, 0, 25, 29],
                        [25, 26, 27, 28, 29],
                    ],
                ],
            ),
        ],
    )


fixture_names = [
    'hexbeam',
    'airplane_ugrid',
    'uniform_ugrid',
    'uniform2d',
    'hybrid',
    'mesh2d',
    'polyhedron',
    'cow_ugrid',
    'empty',
    'points_only',
    'quad_pixel_hex_voxel',
]


@pytest.mark.parametrize('mesh_in_name', fixture_names)
def test_meshio(request, mesh_in_name):
    mesh_in = request.getfixturevalue(mesh_in_name)
    if isinstance(mesh_in, meshio.Mesh):
        mesh_in = pv.from_meshio(mesh_in)

    mesh = pv.from_meshio(pv.to_meshio(mesh_in))

    # Assert mesh is still the same
    assert np.allclose(mesh_in.points, mesh.points)
    if (mesh_in.celltypes == pv.CellType.PIXEL).all():
        cells = mesh_in.cells.reshape((mesh_in.n_cells, 5))[:, [0, 1, 2, 4, 3]].ravel()
        assert np.allclose(cells, mesh.cells)
    elif (mesh_in.celltypes == pv.CellType.VOXEL).all():
        cells = mesh_in.cells.reshape((mesh_in.n_cells, 9))[:, [0, 1, 2, 4, 3, 5, 6, 8, 7]].ravel()
        assert np.allclose(cells, mesh.cells)
    # Mixed cell types with voxels or pixels
    elif np.isin(mesh_in.celltypes, [pv.CellType.PIXEL, pv.CellType.VOXEL]).any():
        cells_in = []

        for cell_type, cell in mesh_in.cells_dict.items():
            for indices in cell:
                # Pixel
                if cell_type == pv.CellType.PIXEL:
                    cells_in.extend([len(indices), *indices[[0, 1, 3, 2]].tolist()])
                # Voxel
                elif cell_type == pv.CellType.VOXEL:
                    cells_in.extend([len(indices), *indices[[0, 1, 3, 2, 4, 5, 7, 6]].tolist()])
                else:
                    cells_in.extend([len(indices), *indices.tolist()])

        cells_out = [
            i
            for cell in mesh.cells_dict.values()
            for indices in cell
            for i in [len(indices), *indices.tolist()]
        ]

        assert cells_in == cells_out
    else:
        assert np.allclose(mesh_in.cells, mesh.cells)
    for k, v in mesh_in.point_data.items():
        assert np.allclose(v, mesh.point_data[k.replace(' ', '_')])
    for k, v in mesh_in.cell_data.items():
        assert np.allclose(v, mesh.cell_data[k.replace(' ', '_')])


def test_pathlib_read_write(tmpdir, sphere):
    path = pathlib.Path(str(tmpdir.mkdir('tmpdir').join('tmp.vtk')))
    pv.save_meshio(path, sphere)
    assert path.is_file()

    mesh = pv.read_meshio(path)
    assert isinstance(mesh, pv.UnstructuredGrid)
    assert mesh.points.shape == sphere.points.shape


def test_file_format(hexbeam):
    from meshio._exceptions import ReadError
    from meshio._exceptions import WriteError

    with pytest.raises(ReadError):
        _ = pv.read_meshio(examples.hexbeamfile, file_format='bar')

    with pytest.raises((KeyError, WriteError)):
        pv.save_meshio('foo.bar', hexbeam, file_format='bar')

    with pytest.raises((KeyError, WriteError)):
        pv.save_meshio('foo.npy', hexbeam, file_format='npy')
